# Copyright 2004-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.cache.mappings import UserDict
from portage.proxy.objectproxy import ObjectProxy
from portage.localization import _
from portage.exception import (
    InvalidAtom,
    PortageException,
    FileNotFound,
    IsADirectory,
    OperationNotPermitted,
    ParseError,
    PermissionDenied,
    ReadOnlyFileSystem,
)
from portage.const import VCS_DIRS
from portage import _unicode_decode
from portage import _unicode_encode
from portage import _os_merge
from portage import _encodings
from portage import os

__all__ = [
    "apply_permissions",
    "apply_recursive_permissions",
    "apply_secpass_permissions",
    "apply_stat_permissions",
    "atomic_ofstream",
    "cmp_sort_key",
    "ConfigProtect",
    "dump_traceback",
    "ensure_dirs",
    "find_updated_config_files",
    "getconfig",
    "getlibpaths",
    "grabdict",
    "grabdict_package",
    "grabfile",
    "grabfile_package",
    "grablines",
    "initialize_logger",
    "LazyItemsDict",
    "map_dictlist_vals",
    "new_protect_filename",
    "normalize_path",
    "pickle_read",
    "stack_dictlist",
    "stack_dicts",
    "stack_lists",
    "unique_array",
    "unique_everseen",
    "varexpand",
    "write_atomic",
    "writedict",
    "writemsg",
    "writemsg_level",
    "writemsg_stdout",
    "no_color",
]

from contextlib import AbstractContextManager
from copy import deepcopy
import errno
import io
from itertools import chain, filterfalse
import logging
import re
import shlex
import stat
import string
import sys
import traceback
import glob
from typing import Optional, TextIO

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "pickle",
    "portage.dep:Atom",
    "subprocess",
)


noiselimit = 0


def initialize_logger(level=logging.WARNING) -> None:
    """Sets up basic logging of portage activities
    Args:
            level: the level to emit messages at ('info', 'debug', 'warning' ...)
    Returns:
            None
    """
    logging.basicConfig(level=level, format="[%(levelname)-4s] %(message)s")


def writemsg(mystr: str, noiselevel: int = 0, fd: Optional[TextIO] = None) -> None:
    """
    Prints out warning and debug messages based on the noiselimit setting

    Takes three arguments
    1. mystr: The message to write
    2. noiselevel: The noiselevel of the message
    3. fd: file descriptor - where to write the message to
    """
    global noiselimit
    if fd is None:
        fd = sys.stderr
    if noiselevel <= noiselimit:
        # avoid potential UnicodeEncodeError
        if isinstance(fd, io.StringIO):
            mystr = _unicode_decode(
                mystr, encoding=_encodings["content"], errors="replace"
            )
        else:
            mystr = _unicode_encode(
                mystr, encoding=_encodings["stdio"], errors="backslashreplace"
            )
            if fd in (sys.stdout, sys.stderr):
                fd = fd.buffer
        fd.write(mystr)
        fd.flush()


def writemsg_stdout(mystr: str, noiselevel: int = 0) -> None:
    """Prints messages stdout based on the noiselimit setting"""
    writemsg(mystr, noiselevel=noiselevel, fd=sys.stdout)


def writemsg_level(msg: str, level: int = 0, noiselevel: int = 0) -> None:
    """
    Show a message for the given level as defined by the logging module
    (default is 0).

    When level >= logging.WARNING then the message is
    sent to stderr, otherwise it is sent to stdout. The noiselevel is
    passed directly to writemsg().
    Takes three parameters
    1. msg  - the message to output
    2. level - the numeric logging level (see python's logging module)
    3. noiselevel - portage's logging level, passed directly to writemsg
    """
    if level >= logging.WARNING:
        fd = sys.stderr
    else:
        fd = sys.stdout
    writemsg(msg, noiselevel=noiselevel, fd=fd)


def normalize_path(mypath) -> str:
    """
    os.path.normpath("//foo") returns "//foo" instead of "/foo"
    We dislike this behavior so we create our own normpath func
    to fix it.
    """
    if isinstance(mypath, bytes):
        path_sep = os.path.sep.encode()
    else:
        path_sep = os.path.sep

    if mypath.startswith(path_sep):
        # posixpath.normpath collapses 3 or more leading slashes to just 1.
        return os.path.normpath(2 * path_sep + mypath)
    return os.path.normpath(mypath)


def grabfile(myfilename, compat_level=0, recursive=0, remember_source_file=False):
    """This function grabs the lines in a file, normalizes whitespace and returns lines in a list; if a line
    begins with a #, it is ignored, as are empty lines"""

    mylines = grablines(myfilename, recursive, remember_source_file=True)
    newlines = []

    for x, source_file in mylines:
        # the split/join thing removes leading and trailing whitespace, and converts any whitespace in the line
        # into single spaces.
        myline = x.split()
        if x and x[0] != "#":
            mylinetemp = []
            for item in myline:
                if item[:1] != "#":
                    mylinetemp.append(item)
                else:
                    break
            myline = mylinetemp

        myline = " ".join(myline)
        if not myline:
            continue
        if myline[0] == "#":
            # Check if we have a compat-level string. BC-integration data.
            # '##COMPAT==>N<==' 'some string attached to it'
            mylinetest = myline.split("<==", 1)
            if len(mylinetest) == 2:
                myline_potential = mylinetest[1]
                mylinetest = mylinetest[0].split("##COMPAT==>")
                if len(mylinetest) == 2:
                    if compat_level >= int(mylinetest[1]):
                        # It's a compat line, and the key matches.
                        newlines.append(myline_potential)
                continue
            else:
                continue
        if remember_source_file:
            newlines.append((myline, source_file))
        else:
            newlines.append(myline)
    return newlines


def map_dictlist_vals(func, myDict):
    """Performs a function on each value of each key in a dictlist.
    Returns a new dictlist."""
    new_dl = {}
    for key in myDict:
        new_dl[key] = []
        new_dl[key] = [func(x) for x in myDict[key]]
    return new_dl


def stack_dictlist(original_dicts, incremental=0, incrementals=[], ignore_none=0):
    """
    Stacks an array of dict-types into one array. Optionally merging or
    overwriting matching key/value pairs for the dict[key]->list.
    Returns a single dict. Higher index in lists is preferenced.

    Example usage:
       >>> from portage.util import stack_dictlist
            >>> print stack_dictlist( [{'a':'b'},{'x':'y'}])
            >>> {'a':'b','x':'y'}
            >>> print stack_dictlist( [{'a':'b'},{'a':'c'}], incremental = True )
            >>> {'a':['b','c'] }
            >>> a = {'KEYWORDS':['x86','alpha']}
            >>> b = {'KEYWORDS':['-x86']}
            >>> print stack_dictlist( [a,b] )
            >>> { 'KEYWORDS':['x86','alpha','-x86']}
            >>> print stack_dictlist( [a,b], incremental=True)
            >>> { 'KEYWORDS':['alpha'] }
            >>> print stack_dictlist( [a,b], incrementals=['KEYWORDS'])
            >>> { 'KEYWORDS':['alpha'] }

    @param original_dicts a list of (dictionary objects or None)
    @type list
    @param incremental True or false depending on whether new keys should overwrite
       keys which already exist.
    @type boolean
    @param incrementals A list of items that should be incremental (-foo removes foo from
       the returned dict).
    @type list
    @param ignore_none Appears to be ignored, but probably was used long long ago.
    @type boolean

    """
    final_dict = {}
    for mydict in original_dicts:
        if mydict is None:
            continue
        for y in mydict:
            if not y in final_dict:
                final_dict[y] = []

            for thing in mydict[y]:
                if thing:
                    if incremental or y in incrementals:
                        if thing == "-*":
                            final_dict[y] = []
                            continue
                        elif thing[:1] == "-":
                            try:
                                final_dict[y].remove(thing[1:])
                            except ValueError:
                                pass
                            continue
                    if thing not in final_dict[y]:
                        final_dict[y].append(thing)
            if y in final_dict and not final_dict[y]:
                del final_dict[y]
    return final_dict


def stack_dicts(dicts, incremental=0, incrementals=[], ignore_none=0):
    """Stacks an array of dict-types into one array. Optionally merging or
    overwriting matching key/value pairs for the dict[key]->string.
    Returns a single dict."""
    final_dict = {}
    for mydict in dicts:
        if not mydict:
            continue
        for k, v in mydict.items():
            if k in final_dict and (incremental or (k in incrementals)):
                final_dict[k] += " " + v
            else:
                final_dict[k] = v
    return final_dict


def append_repo(atom_list, repo_name, remember_source_file=False):
    """
    Takes a list of valid atoms without repo spec and appends ::repo_name.
    If an atom already has a repo part, then it is preserved (see bug #461948).
    """
    if remember_source_file:
        return [
            (atom.repo is not None and atom or atom.with_repo(repo_name), source)
            for atom, source in atom_list
        ]

    return [
        atom.repo is not None and atom or atom.with_repo(repo_name)
        for atom in atom_list
    ]


def stack_lists(
    lists,
    incremental=1,
    remember_source_file=False,
    warn_for_unmatched_removal=False,
    strict_warn_for_unmatched_removal=False,
    ignore_repo=False,
):
    """Stacks an array of list-types into one array. Optionally removing
    distinct values using '-value' notation. Higher index is preferenced.

    all elements must be hashable."""
    matched_removals = set()
    unmatched_removals = {}
    new_list = {}
    for sub_list in lists:
        for token in sub_list:
            token_key = token
            if remember_source_file:
                token, source_file = token
            else:
                source_file = False

            if token is None:
                continue

            if incremental:
                if token == "-*":
                    new_list.clear()
                elif token[:1] == "-":
                    matched = False
                    if ignore_repo and not "::" in token:
                        # Let -cat/pkg remove cat/pkg::repo.
                        to_be_removed = []
                        token_slice = token[1:]
                        for atom in new_list:
                            atom_without_repo = atom
                            if atom.repo is not None:
                                # Atom.without_repo instantiates a new Atom,
                                # which is unnecessary here, so use string
                                # replacement instead.
                                atom_without_repo = atom.replace(
                                    "::" + atom.repo, "", 1
                                )
                            if atom_without_repo == token_slice:
                                to_be_removed.append(atom)
                        if to_be_removed:
                            matched = True
                            for atom in to_be_removed:
                                new_list.pop(atom)
                    else:
                        try:
                            new_list.pop(token[1:])
                            matched = True
                        except KeyError:
                            pass

                    if not matched:
                        if source_file and (
                            strict_warn_for_unmatched_removal
                            or token_key not in matched_removals
                        ):
                            unmatched_removals.setdefault(source_file, set()).add(token)
                    else:
                        matched_removals.add(token_key)
                else:
                    new_list[token] = source_file
            else:
                new_list[token] = source_file

    if warn_for_unmatched_removal:
        for source_file, tokens in unmatched_removals.items():
            if len(tokens) > 3:
                selected = [tokens.pop(), tokens.pop(), tokens.pop()]
                writemsg(
                    _("--- Unmatched removal atoms in %s: %s and %s more\n")
                    % (source_file, ", ".join(selected), len(tokens)),
                    noiselevel=-1,
                )
            else:
                writemsg(
                    _("--- Unmatched removal atom(s) in %s: %s\n")
                    % (source_file, ", ".join(tokens)),
                    noiselevel=-1,
                )

    if remember_source_file:
        return list(new_list.items())
    return list(new_list)


def grabdict(
    myfilename, juststrings=0, empty=0, recursive=0, incremental=1, newlines=0
):
    """
    This function grabs the lines in a file, normalizes whitespace and returns lines in a dictionary

    @param myfilename: file to process
    @type myfilename: string (path)
    @param juststrings: only return strings
    @type juststrings: Boolean (integer)
    @param empty: Ignore certain lines
    @type empty: Boolean (integer)
    @param recursive: Recursively grab ( support for /etc/portage/package.keywords/* and friends )
    @type recursive: Boolean (integer)
    @param incremental: Append to the return list, don't overwrite
    @type incremental: Boolean (integer)
    @param newlines: Append newlines
    @type newlines: Boolean (integer)
    @rtype: Dictionary
    @return:
    1.  Returns the lines in a file in a dictionary, for example:
            'sys-apps/portage x86 amd64 ppc'
            would return
            {"sys-apps/portage" : ['x86', 'amd64', 'ppc']}
    """
    newdict = {}
    for x in grablines(myfilename, recursive):
        # the split/join thing removes leading and trailing whitespace, and converts any whitespace in the line
        # into single spaces.
        if x[0] == "#":
            continue
        myline = x.split()
        mylinetemp = []
        for item in myline:
            if item[:1] != "#":
                mylinetemp.append(item)
            else:
                break
        myline = mylinetemp
        if len(myline) < 2 and empty == 0:
            continue
        if len(myline) < 1 and empty == 1:
            continue
        if newlines:
            myline.append("\n")
        if incremental:
            newdict.setdefault(myline[0], []).extend(myline[1:])
        else:
            newdict[myline[0]] = myline[1:]
    if juststrings:
        for k, v in newdict.items():
            newdict[k] = " ".join(v)
    return newdict


_eapi_cache = {}


def read_corresponding_eapi_file(filename, default="0"):
    """
    Read the 'eapi' file from the directory 'filename' is in.
    Returns "0" if the file is not present or invalid.
    """
    eapi_file = os.path.join(os.path.dirname(filename), "eapi")
    try:
        eapi = _eapi_cache[eapi_file]
    except KeyError:
        pass
    else:
        if eapi is None:
            return default
        return eapi

    eapi = None
    try:
        with open(
            _unicode_encode(eapi_file, encoding=_encodings["fs"], errors="strict"),
            encoding=_encodings["repo.content"],
            errors="replace",
        ) as f:
            lines = f.readlines()
        if len(lines) == 1:
            eapi = lines[0].rstrip("\n")
        else:
            writemsg(
                _("--- Invalid 'eapi' file (doesn't contain exactly one line): %s\n")
                % (eapi_file),
                noiselevel=-1,
            )
    except OSError:
        pass

    _eapi_cache[eapi_file] = eapi
    if eapi is None:
        return default
    return eapi


def grabdict_package(
    myfilename,
    juststrings=0,
    recursive=0,
    newlines=0,
    allow_wildcard=False,
    allow_repo=False,
    allow_build_id=False,
    allow_use=True,
    verify_eapi=False,
    eapi=None,
    eapi_default="0",
):
    """Does the same thing as grabdict except it validates keys
    with isvalidatom()"""

    if recursive:
        file_list = _recursive_file_list(myfilename)
    else:
        file_list = [myfilename]

    atoms = {}
    for filename in file_list:
        d = grabdict(
            filename,
            juststrings=False,
            empty=True,
            recursive=False,
            incremental=True,
            newlines=newlines,
        )
        if not d:
            continue
        if verify_eapi and eapi is None:
            eapi = read_corresponding_eapi_file(myfilename, default=eapi_default)

        for k, v in d.items():
            try:
                k = Atom(
                    k,
                    allow_wildcard=allow_wildcard,
                    allow_repo=allow_repo,
                    allow_build_id=allow_build_id,
                    eapi=eapi,
                )
            except InvalidAtom as e:
                writemsg(
                    _("--- Invalid atom in %s: %s\n") % (filename, e), noiselevel=-1
                )
            else:
                if not allow_use and k.use:
                    writemsg(
                        _("--- Atom is not allowed to have USE flag(s) in %s: %s\n")
                        % (filename, k),
                        noiselevel=-1,
                    )
                    continue
                atoms.setdefault(k, []).extend(v)

    if juststrings:
        for k, v in atoms.items():
            atoms[k] = " ".join(v)

    return atoms


def grabfile_package(
    myfilename,
    compatlevel=0,
    recursive=0,
    allow_wildcard=False,
    allow_repo=False,
    allow_build_id=False,
    remember_source_file=False,
    verify_eapi=False,
    eapi=None,
    eapi_default="0",
):
    pkgs = grabfile(
        myfilename, compatlevel, recursive=recursive, remember_source_file=True
    )
    if not pkgs:
        return pkgs
    if verify_eapi and eapi is None:
        eapi = read_corresponding_eapi_file(myfilename, default=eapi_default)
    mybasename = os.path.basename(myfilename)
    is_packages_file = mybasename == "packages"
    atoms = []
    for pkg, source_file in pkgs:
        pkg_orig = pkg
        # for packages and package.mask files
        if pkg[:1] == "-":
            if is_packages_file and pkg == "-*":
                if remember_source_file:
                    atoms.append((pkg, source_file))
                else:
                    atoms.append(pkg)
                continue
            pkg = pkg[1:]
        if pkg[:1] == "*" and is_packages_file:
            pkg = pkg[1:]
        try:
            pkg = Atom(
                pkg,
                allow_wildcard=allow_wildcard,
                allow_repo=allow_repo,
                allow_build_id=allow_build_id,
                eapi=eapi,
            )
        except InvalidAtom as e:
            writemsg(
                _("--- Invalid atom in %s: %s\n") % (source_file, e), noiselevel=-1
            )
        else:
            if pkg_orig == str(pkg):
                # normal atom, so return as Atom instance
                if remember_source_file:
                    atoms.append((pkg, source_file))
                else:
                    atoms.append(pkg)
            else:
                # atom has special prefix, so return as string
                if remember_source_file:
                    atoms.append((pkg_orig, source_file))
                else:
                    atoms.append(pkg_orig)
    return atoms


def _recursive_basename_filter(f):
    return not f.startswith(".") and not f.endswith("~")


def _recursive_file_list(path):
    # path may be a regular file or a directory

    def onerror(e):
        if e.errno == PermissionDenied.errno:
            raise PermissionDenied(path)

    stack = [os.path.split(path)]

    while stack:
        parent, fname = stack.pop()
        fullpath = os.path.join(parent, fname)

        try:
            st = os.stat(fullpath)
        except OSError as e:
            onerror(e)
            continue

        if stat.S_ISDIR(st.st_mode):
            if fname in VCS_DIRS or not _recursive_basename_filter(fname):
                continue
            try:
                children = os.listdir(fullpath)
            except OSError as e:
                onerror(e)
                continue

            # Sort in reverse, since we pop from the end of the stack.
            # Include regular files in the stack, so files are sorted
            # together with directories.
            children.sort(reverse=True)
            stack.extend((fullpath, x) for x in children)

        elif stat.S_ISREG(st.st_mode):
            if _recursive_basename_filter(fname):
                yield fullpath


def grablines(myfilename, recursive=0, remember_source_file=False):
    mylines = []
    if recursive:
        for f in _recursive_file_list(myfilename):
            mylines.extend(
                grablines(f, recursive=False, remember_source_file=remember_source_file)
            )

    else:
        try:
            with open(
                _unicode_encode(myfilename, encoding=_encodings["fs"], errors="strict"),
                encoding=_encodings["content"],
                errors="replace",
            ) as myfile:
                if remember_source_file:
                    mylines = [(line, myfilename) for line in myfile.readlines()]
                else:
                    mylines = myfile.readlines()
        except OSError as e:
            if e.errno == PermissionDenied.errno:
                raise PermissionDenied(myfilename)
            elif e.errno in (errno.ENOENT, errno.ESTALE):
                pass
            else:
                raise
    return mylines


def writedict(mydict, myfilename, writekey=True):
    """Writes out a dict to a file; writekey=0 mode doesn't write out
    the key and assumes all values are strings, not lists."""
    lines = []
    if not writekey:
        for v in mydict.values():
            lines.append(v + "\n")
    else:
        for k, v in mydict.items():
            lines.append(f"{k} {' '.join(v)}\n")
    write_atomic(myfilename, "".join(lines))


def shlex_split(s):
    """
    This is equivalent to shlex.split, but if the current interpreter is
    python2, it temporarily encodes unicode strings to bytes since python2's
    shlex.split() doesn't handle unicode strings.
    """
    return shlex.split(s)


class _getconfig_shlex(shlex.shlex):
    def __init__(self, portage_tolerant=False, **kwargs):
        shlex.shlex.__init__(self, **kwargs)
        self.__portage_tolerant = portage_tolerant

    def allow_sourcing(self, var_expand_map):
        self.source = portage._native_string("source")
        self.var_expand_map = var_expand_map

    def sourcehook(self, newfile):
        try:
            newfile = varexpand(newfile, self.var_expand_map)
            return shlex.shlex.sourcehook(self, newfile)
        except OSError as e:
            if e.errno == PermissionDenied.errno:
                raise PermissionDenied(newfile)
            if e.errno not in (errno.ENOENT, errno.ENOTDIR):
                writemsg(f"open('{newfile}', 'r'): {e}\n", noiselevel=-1)
                raise

            msg = self.error_leader()
            if e.errno == errno.ENOTDIR:
                msg += _("%s: Not a directory") % newfile
            else:
                msg += _("%s: No such file or directory") % newfile

            if self.__portage_tolerant:
                writemsg(f"{msg}\n", noiselevel=-1)
            else:
                raise ParseError(msg)
            return (newfile, io.StringIO())


_invalid_var_name_re = re.compile(r"^\d|\W")


def getconfig(
    mycfg, tolerant=False, allow_sourcing=False, expand=True, recursive=False
):
    if isinstance(expand, dict):
        # Some existing variable definitions have been
        # passed in, for use in substitutions.
        expand_map = expand
        expand = True
    else:
        expand_map = {}
    mykeys = {}

    if recursive:
        # Emulate source commands so that syntax error messages
        # can display real file names and line numbers.
        if not expand:
            expand_map = False
        fname = None
        for fname in _recursive_file_list(mycfg):
            mykeys.update(
                getconfig(
                    fname,
                    tolerant=tolerant,
                    allow_sourcing=allow_sourcing,
                    expand=expand_map,
                    recursive=False,
                )
                or {}
            )
        if fname is None:
            return None
        return mykeys

    f = None
    try:
        f = open(
            _unicode_encode(mycfg, encoding=_encodings["fs"], errors="strict"),
            encoding=_encodings["content"],
            errors="replace",
        )
        content = f.read()
    except OSError as e:
        if e.errno == PermissionDenied.errno:
            raise PermissionDenied(mycfg)
        if e.errno != errno.ENOENT:
            writemsg(f"open('{mycfg}', 'r'): {e}\n", noiselevel=-1)
            if e.errno not in (errno.EISDIR,):
                raise
        return None
    finally:
        if f is not None:
            f.close()

    # Since this file has unicode_literals enabled, and Python 2's
    # shlex implementation does not support unicode, the following code
    # uses _native_string() to encode unicode literals when necessary.

    # Workaround for avoiding a silent error in shlex that is
    # triggered by a source statement at the end of the file
    # without a trailing newline after the source statement.
    if content and content[-1] != portage._native_string("\n"):
        content += portage._native_string("\n")

    # Warn about dos-style line endings since that prevents
    # people from being able to source them with bash.
    if portage._native_string("\r") in content:
        writemsg(
            (
                "!!! "
                + _(
                    "Please use dos2unix to convert line endings "
                    + "in config file: '%s'"
                )
                + "\n"
            )
            % mycfg,
            noiselevel=-1,
        )

    lex = None
    try:
        # The default shlex.sourcehook() implementation
        # only joins relative paths when the infile
        # attribute is properly set.
        lex = _getconfig_shlex(
            instream=content, infile=mycfg, posix=True, portage_tolerant=tolerant
        )
        lex.wordchars = portage._native_string(
            string.digits + string.ascii_letters + r"~!@#$%*_\:;?,./-+{}"
        )
        lex.quotes = portage._native_string("\"'")
        if allow_sourcing:
            lex.allow_sourcing(expand_map)

        while True:
            key = _unicode_decode(lex.get_token())
            if key == "export":
                key = _unicode_decode(lex.get_token())
            if key is None:
                # normal end of file
                break

            equ = _unicode_decode(lex.get_token())
            if not equ:
                msg = lex.error_leader() + _("Unexpected EOF")
                if not tolerant:
                    raise ParseError(msg)
                else:
                    writemsg(f"{msg}\n", noiselevel=-1)
                    return mykeys

            elif equ != "=":
                msg = lex.error_leader() + _("Invalid token '%s' (not '=')") % (equ,)
                if not tolerant:
                    raise ParseError(msg)
                else:
                    writemsg(f"{msg}\n", noiselevel=-1)
                    return mykeys

            val = _unicode_decode(lex.get_token())
            if val is None:
                msg = lex.error_leader() + _(
                    "Unexpected end of config file: variable '%s'"
                ) % (key,)
                if not tolerant:
                    raise ParseError(msg)
                else:
                    writemsg(f"{msg}\n", noiselevel=-1)
                    return mykeys

            if _invalid_var_name_re.search(key) is not None:
                msg = lex.error_leader() + _("Invalid variable name '%s'") % (key,)
                if not tolerant:
                    raise ParseError(msg)
                writemsg(f"{msg}\n", noiselevel=-1)
                continue

            if expand:
                mykeys[key] = varexpand(
                    val, mydict=expand_map, error_leader=lex.error_leader
                )
                expand_map[key] = mykeys[key]
            else:
                mykeys[key] = val
    except SystemExit as e:
        raise
    except Exception as e:
        if isinstance(e, ParseError) or lex is None:
            raise
        msg = f"{lex.error_leader()}{e}"
        writemsg(f"{msg}\n", noiselevel=-1)
        raise

    return mykeys


_varexpand_word_chars = frozenset(string.ascii_letters + string.digits + "_")
_varexpand_unexpected_eof_msg = "unexpected EOF while looking for matching `}'"


def varexpand(mystring, mydict=None, error_leader=None):
    if mydict is None:
        mydict = {}

    """
	new variable expansion code.  Preserves quotes, handles \n, etc.
	This code is used by the configfile code, as well as others (parser)
	This would be a good bunch of code to port to C.
	"""
    numvars = 0
    # in single, double quotes
    insing = 0
    indoub = 0
    pos = 0
    length = len(mystring)
    newstring = []
    while pos < length:
        current = mystring[pos]
        if current == "'":
            if indoub:
                newstring.append("'")
            else:
                newstring.append("'")  # Quote removal is handled by shlex.
                insing = not insing
            pos += 1
            continue
        elif current == '"':
            if insing:
                newstring.append('"')
            else:
                newstring.append('"')  # Quote removal is handled by shlex.
                indoub = not indoub
            pos += 1
            continue
        if not insing:
            # expansion time
            if current == "\n":
                # convert newlines to spaces
                newstring.append(" ")
                pos += 1
            elif current == "\\":
                # For backslash expansion, this function used to behave like
                # echo -e, but that's not needed for our purposes. We want to
                # behave like bash does when expanding a variable assignment
                # in a sourced file, in which case it performs backslash
                # removal for \\ and \$ but nothing more. It also removes
                # escaped newline characters. Note that we don't handle
                # escaped quotes here, since getconfig() uses shlex
                # to handle that earlier.
                if pos + 1 >= len(mystring):
                    newstring.append(current)
                    break
                else:
                    current = mystring[pos + 1]
                    pos += 2
                    if current == "$":
                        newstring.append(current)
                    elif current == "\\":
                        newstring.append(current)
                        # BUG: This spot appears buggy, but it's intended to
                        # be bug-for-bug compatible with existing behavior.
                        if pos < length and mystring[pos] in ("'", '"', "$"):
                            newstring.append(mystring[pos])
                            pos += 1
                    elif current == "\n":
                        pass
                    else:
                        newstring.append(mystring[pos - 2 : pos])
                    continue
            elif current == "$":
                pos += 1
                if pos == length:
                    # shells handle this like \$
                    newstring.append(current)
                    continue

                if mystring[pos] == "{":
                    pos += 1
                    if pos == length:
                        msg = _varexpand_unexpected_eof_msg
                        if error_leader is not None:
                            msg = error_leader() + msg
                        writemsg(msg + "\n", noiselevel=-1)
                        return ""

                    braced = True
                else:
                    braced = False
                myvstart = pos
                while mystring[pos] in _varexpand_word_chars:
                    if pos + 1 >= len(mystring):
                        if braced:
                            msg = _varexpand_unexpected_eof_msg
                            if error_leader is not None:
                                msg = error_leader() + msg
                            writemsg(msg + "\n", noiselevel=-1)
                            return ""
                        pos += 1
                        break
                    pos += 1
                myvarname = mystring[myvstart:pos]
                if braced:
                    if mystring[pos] != "}":
                        msg = _varexpand_unexpected_eof_msg
                        if error_leader is not None:
                            msg = error_leader() + msg
                        writemsg(msg + "\n", noiselevel=-1)
                        return ""
                    pos += 1
                if len(myvarname) == 0:
                    msg = "$"
                    if braced:
                        msg += "{}"
                    msg += ": bad substitution"
                    if error_leader is not None:
                        msg = error_leader() + msg
                    writemsg(msg + "\n", noiselevel=-1)
                    return ""
                numvars += 1
                if myvarname in mydict:
                    newstring.append(mydict[myvarname])
            else:
                newstring.append(current)
                pos += 1
        else:
            newstring.append(current)
            pos += 1

    return "".join(newstring)


# broken and removed, but can still be imported
pickle_write = None


def pickle_read(filename, default=None, debug=0):
    if not os.access(filename, os.R_OK):
        writemsg(_("pickle_read(): File not readable. '") + filename + "'\n", 1)
        return default
    data = None
    try:
        myf = open(
            _unicode_encode(filename, encoding=_encodings["fs"], errors="strict"), "rb"
        )
        mypickle = pickle.Unpickler(myf)
        data = mypickle.load()
        myf.close()
        del mypickle, myf
        writemsg(_("pickle_read(): Loaded pickle. '") + filename + "'\n", 1)
    except SystemExit as e:
        raise
    except Exception as e:
        writemsg(_("!!! Failed to load pickle: ") + str(e) + "\n", 1)
        data = default
    return data


def dump_traceback(msg, noiselevel=1):
    info = sys.exc_info()
    if not info[2]:
        stack = traceback.extract_stack()[:-1]
        error = None
    else:
        stack = traceback.extract_tb(info[2])
        error = str(info[1])
    writemsg("\n====================================\n", noiselevel=noiselevel)
    writemsg(f"{msg}\n\n", noiselevel=noiselevel)
    for line in traceback.format_list(stack):
        writemsg(line, noiselevel=noiselevel)
    if error:
        writemsg(error + "\n", noiselevel=noiselevel)
    writemsg("====================================\n\n", noiselevel=noiselevel)


class cmp_sort_key:
    """
    In python-3.0 the list.sort() method no longer has a "cmp" keyword
    argument. This class acts as an adapter which converts a cmp function
    into one that's suitable for use as the "key" keyword argument to
    list.sort(), making it easier to port code for python-3.0 compatibility.
    It works by generating key objects which use the given cmp function to
    implement their __lt__ method.

    Beginning with Python 2.7 and 3.2, equivalent functionality is provided
    by functools.cmp_to_key().
    """

    __slots__ = ("_cmp_func",)

    def __init__(self, cmp_func):
        """
        @type cmp_func: callable which takes 2 positional arguments
        @param cmp_func: A cmp function.
        """
        self._cmp_func = cmp_func

    def __call__(self, lhs):
        return self._cmp_key(self._cmp_func, lhs)

    class _cmp_key:
        __slots__ = ("_cmp_func", "_obj")

        def __init__(self, cmp_func, obj):
            self._cmp_func = cmp_func
            self._obj = obj

        def __lt__(self, other):
            if other.__class__ is not self.__class__:
                raise TypeError(
                    f"Expected type {self.__class__}, got {other.__class__}"
                )
            return self._cmp_func(self._obj, other._obj) < 0


def unique_array(s):
    """lifted from python cookbook, credit: Tim Peters
    Return a list of the elements in s in arbitrary order, sans duplicates"""
    n = len(s)
    # assume all elements are hashable, if so, it's linear
    try:
        return list(set(s))
    except TypeError:
        pass

    # so much for linear.  abuse sort.
    try:
        t = list(s)
        t.sort()
    except TypeError:
        pass
    else:
        assert n > 0
        last = t[0]
        lasti = i = 1
        while i < n:
            if t[i] != last:
                t[lasti] = last = t[i]
                lasti += 1
            i += 1
        return t[:lasti]

    # blah.	 back to original portage.unique_array
    u = []
    for x in s:
        if x not in u:
            u.append(x)
    return u


def unique_everseen(iterable, key=None):
    """
    List unique elements, preserving order. Remember all elements ever seen.
    Taken from itertools documentation.
    """
    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
    # unique_everseen('ABBCcAD', str.lower) --> A B C D
    seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element


def _do_stat(filename, follow_links=True):
    try:
        if follow_links:
            return os.stat(filename)
        return os.lstat(filename)
    except OSError as oe:
        func_call = f"stat('{filename}')"
        if oe.errno == errno.EPERM:
            raise OperationNotPermitted(func_call)
        if oe.errno == errno.EACCES:
            raise PermissionDenied(func_call)
        if oe.errno == errno.ENOENT:
            raise FileNotFound(filename)
        raise


def apply_permissions(
    filename, uid=-1, gid=-1, mode=-1, mask=-1, stat_cached=None, follow_links=True
):
    """Apply user, group, and mode bits to a file if the existing bits do not
    already match.  The default behavior is to force an exact match of mode
    bits.  When mask=0 is specified, mode bits on the target file are allowed
    to be a superset of the mode argument (via logical OR).  When mask>0, the
    mode bits that the target file is allowed to have are restricted via
    logical XOR.
    Returns True if the permissions were modified and False otherwise."""

    modified = False

    # Since Python 3.4, chown requires int type (no proxies).
    uid = int(uid)
    gid = int(gid)

    if stat_cached is None:
        stat_cached = _do_stat(filename, follow_links=follow_links)

    if (uid != -1 and uid != stat_cached.st_uid) or (
        gid != -1 and gid != stat_cached.st_gid
    ):
        try:
            if follow_links:
                os.chown(filename, uid, gid)
            else:
                portage.data.lchown(filename, uid, gid)
            modified = True
        except OSError as oe:
            func_call = "chown('%s', %i, %i)" % (filename, uid, gid)
            if oe.errno == errno.EPERM:
                raise OperationNotPermitted(func_call)
            elif oe.errno == errno.EACCES:
                raise PermissionDenied(func_call)
            elif oe.errno == errno.EROFS:
                raise ReadOnlyFileSystem(func_call)
            elif oe.errno == errno.ENOENT:
                raise FileNotFound(filename)
            else:
                raise

    new_mode = -1
    st_mode = stat_cached.st_mode & 0o7777  # protect from unwanted bits
    if mask >= 0:
        if mode == -1:
            mode = 0  # Don't add any mode bits when mode is unspecified.
        else:
            mode = mode & 0o7777
        if (mode & st_mode != mode) or ((mask ^ st_mode) & st_mode != st_mode):
            new_mode = mode | st_mode
            new_mode = (mask ^ new_mode) & new_mode
    elif mode != -1:
        mode = mode & 0o7777  # protect from unwanted bits
        if mode != st_mode:
            new_mode = mode

    # The chown system call may clear S_ISUID and S_ISGID
    # bits, so those bits are restored if necessary.
    if (
        modified
        and new_mode == -1
        and (st_mode & stat.S_ISUID or st_mode & stat.S_ISGID)
    ):
        if mode == -1:
            new_mode = st_mode
        else:
            mode = mode & 0o7777
            if mask >= 0:
                new_mode = mode | st_mode
                new_mode = (mask ^ new_mode) & new_mode
            else:
                new_mode = mode
            if not (new_mode & stat.S_ISUID or new_mode & stat.S_ISGID):
                new_mode = -1

    if not follow_links and stat.S_ISLNK(stat_cached.st_mode):
        # Mode doesn't matter for symlinks.
        new_mode = -1

    if new_mode != -1:
        try:
            os.chmod(filename, new_mode)
            modified = True
        except OSError as oe:
            func_call = f"chmod('{filename}', {oct(new_mode)})"
            if oe.errno == errno.EPERM:
                raise OperationNotPermitted(func_call)
            elif oe.errno == errno.EACCES:
                raise PermissionDenied(func_call)
            elif oe.errno == errno.EROFS:
                raise ReadOnlyFileSystem(func_call)
            elif oe.errno == errno.ENOENT:
                raise FileNotFound(filename)
            raise
    return modified


def apply_stat_permissions(filename, newstat, **kwargs):
    """A wrapper around apply_secpass_permissions that gets
    uid, gid, and mode from a stat object"""
    return apply_secpass_permissions(
        filename, uid=newstat.st_uid, gid=newstat.st_gid, mode=newstat.st_mode, **kwargs
    )


def apply_recursive_permissions(
    top, uid=-1, gid=-1, dirmode=-1, dirmask=-1, filemode=-1, filemask=-1, onerror=None
):
    """A wrapper around apply_secpass_permissions that applies permissions
    recursively.  If optional argument onerror is specified, it should be a
    function; it will be called with one argument, a PortageException instance.
    Returns True if all permissions are applied and False if some are left
    unapplied."""

    # Avoid issues with circular symbolic links, as in bug #339670.
    follow_links = False

    if onerror is None:
        # Default behavior is to dump errors to stderr so they won't
        # go unnoticed.  Callers can pass in a quiet instance.
        def onerror(e):
            if isinstance(e, OperationNotPermitted):
                writemsg(_("Operation Not Permitted: %s\n") % str(e), noiselevel=-1)
            elif isinstance(e, FileNotFound):
                writemsg(_("File Not Found: '%s'\n") % str(e), noiselevel=-1)
            else:
                raise

    # For bug 554084, always apply permissions to a directory before
    # that directory is traversed.
    all_applied = True

    try:
        stat_cached = _do_stat(top, follow_links=follow_links)
    except FileNotFound:
        # backward compatibility
        return True

    if stat.S_ISDIR(stat_cached.st_mode):
        mode = dirmode
        mask = dirmask
    else:
        mode = filemode
        mask = filemask

    try:
        applied = apply_secpass_permissions(
            top,
            uid=uid,
            gid=gid,
            mode=mode,
            mask=mask,
            stat_cached=stat_cached,
            follow_links=follow_links,
        )
        if not applied:
            all_applied = False
    except PortageException as e:
        all_applied = False
        onerror(e)

    for dirpath, dirnames, filenames in os.walk(top):
        for name, mode, mask in chain(
            ((x, filemode, filemask) for x in filenames),
            ((x, dirmode, dirmask) for x in dirnames),
        ):
            try:
                applied = apply_secpass_permissions(
                    os.path.join(dirpath, name),
                    uid=uid,
                    gid=gid,
                    mode=mode,
                    mask=mask,
                    follow_links=follow_links,
                )
                if not applied:
                    all_applied = False
            except PortageException as e:
                # Ignore InvalidLocation exceptions such as FileNotFound
                # and DirectoryNotFound since sometimes things disappear,
                # like when adjusting permissions on DISTCC_DIR.
                if not isinstance(e, portage.exception.InvalidLocation):
                    all_applied = False
                    onerror(e)
    return all_applied


def apply_secpass_permissions(
    filename, uid=-1, gid=-1, mode=-1, mask=-1, stat_cached=None, follow_links=True
):
    """A wrapper around apply_permissions that uses secpass and simple
    logic to apply as much of the permissions as possible without
    generating an obviously avoidable permission exception. Despite
    attempts to avoid an exception, it's possible that one will be raised
    anyway, so be prepared.
    Returns True if all permissions are applied and False if some are left
    unapplied."""

    if stat_cached is None:
        stat_cached = _do_stat(filename, follow_links=follow_links)

    all_applied = True

    # Avoid accessing portage.data.secpass when possible, since
    # it triggers config loading (undesirable for chmod-lite).
    if (uid != -1 or gid != -1) and portage.data.secpass < 2:
        if uid != -1 and uid != stat_cached.st_uid:
            all_applied = False
            uid = -1

        if gid != -1 and gid != stat_cached.st_gid and gid not in os.getgroups():
            all_applied = False
            gid = -1

    apply_permissions(
        filename,
        uid=uid,
        gid=gid,
        mode=mode,
        mask=mask,
        stat_cached=stat_cached,
        follow_links=follow_links,
    )
    return all_applied


class atomic_ofstream(AbstractContextManager, ObjectProxy):
    """Write a file atomically via os.rename().  Atomic replacement prevents
    interprocess interference and prevents corruption of the target
    file when the write is interrupted (for example, when an 'out of space'
    error occurs)."""

    def __init__(self, filename, mode="w", follow_links=True, **kargs):
        """Opens a temporary filename.pid in the same directory as filename."""
        ObjectProxy.__init__(self)
        object.__setattr__(self, "_aborted", False)
        if "b" in mode:
            open_func = open
        else:
            open_func = io.open
            kargs.setdefault("encoding", _encodings["content"])
            kargs.setdefault("errors", "backslashreplace")

        if follow_links:
            canonical_path = os.path.realpath(filename)
            object.__setattr__(self, "_real_name", canonical_path)
            tmp_name = "%s.%i" % (canonical_path, portage.getpid())
            try:
                object.__setattr__(
                    self,
                    "_file",
                    open_func(
                        _unicode_encode(
                            tmp_name, encoding=_encodings["fs"], errors="strict"
                        ),
                        mode=mode,
                        **kargs,
                    ),
                )
                return
            except OSError as e:
                if canonical_path == filename:
                    raise
                # Ignore this error, since it's irrelevant
                # and the below open call will produce a
                # new error if necessary.

        object.__setattr__(self, "_real_name", filename)
        tmp_name = "%s.%i" % (filename, portage.getpid())
        object.__setattr__(
            self,
            "_file",
            open_func(
                _unicode_encode(tmp_name, encoding=_encodings["fs"], errors="strict"),
                mode=mode,
                **kargs,
            ),
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.abort()
        else:
            self.close()

    def _get_target(self):
        return object.__getattribute__(self, "_file")

    def __getattribute__(self, attr):
        if attr in ("close", "abort", "__del__"):
            return object.__getattribute__(self, attr)
        return getattr(object.__getattribute__(self, "_file"), attr)

    def close(self):
        """Closes the temporary file, copies permissions (if possible),
        and performs the atomic replacement via os.rename().  If the abort()
        method has been called, then the temp file is closed and removed."""
        f = object.__getattribute__(self, "_file")
        real_name = object.__getattribute__(self, "_real_name")
        if not f.closed:
            try:
                f.close()
                if not object.__getattribute__(self, "_aborted"):
                    try:
                        apply_stat_permissions(f.name, os.stat(real_name))
                    except OperationNotPermitted:
                        pass
                    except FileNotFound:
                        pass
                    except OSError as oe:  # from the above os.stat call
                        if oe.errno in (errno.ENOENT, errno.EPERM):
                            pass
                        else:
                            raise
                    os.rename(f.name, real_name)
            finally:
                # Make sure we cleanup the temp file
                # even if an exception is raised.
                try:
                    os.unlink(f.name)
                except OSError as oe:
                    pass

    def abort(self):
        """If an error occurs while writing the file, the user should
        call this method in order to leave the target file unchanged.
        This will call close() automatically."""
        if not object.__getattribute__(self, "_aborted"):
            object.__setattr__(self, "_aborted", True)
            self.close()

    def __del__(self):
        """If the user does not explicitly call close(), it is
        assumed that an error has occurred, so we abort()."""
        try:
            f = object.__getattribute__(self, "_file")
        except AttributeError:
            pass
        else:
            if not f.closed:
                self.abort()
        # ensure destructor from the base class is called
        base_destructor = getattr(ObjectProxy, "__del__", None)
        if base_destructor is not None:
            base_destructor(self)


def write_atomic(file_path, content, **kwargs):
    f = None
    try:
        f = atomic_ofstream(file_path, **kwargs)
        f.write(content)
        f.close()
    except OSError as e:
        if f:
            f.abort()
        func_call = f"write_atomic('{file_path}')"
        if e.errno == errno.EPERM:
            raise OperationNotPermitted(func_call)
        elif e.errno == errno.EACCES:
            raise PermissionDenied(func_call)
        elif e.errno == errno.EROFS:
            raise ReadOnlyFileSystem(func_call)
        elif e.errno == errno.ENOENT:
            raise FileNotFound(file_path)
        else:
            raise


def ensure_dirs(dir_path, **kwargs):
    """Create a directory and call apply_permissions.
    Returns True if a directory is created or the permissions needed to be
    modified, and False otherwise.

    This function's handling of EEXIST errors makes it useful for atomic
    directory creation, in which multiple processes may be competing to
    create the same directory.
    """

    created_dir = False

    try:
        os.makedirs(dir_path)
        created_dir = True
    except OSError as oe:
        func_call = f"makedirs('{dir_path}')"
        if oe.errno in (errno.EEXIST,):
            pass
        else:
            if os.path.isdir(dir_path):
                # NOTE: DragonFly raises EPERM for makedir('/')
                # and that is supposed to be ignored here.
                # Also, sometimes mkdir raises EISDIR on FreeBSD
                # and we want to ignore that too (bug #187518).
                pass
            elif oe.errno == errno.EPERM:
                raise OperationNotPermitted(func_call)
            elif oe.errno == errno.EACCES:
                raise PermissionDenied(func_call)
            elif oe.errno == errno.EROFS:
                raise ReadOnlyFileSystem(func_call)
            else:
                raise
    if kwargs:
        perms_modified = apply_permissions(dir_path, **kwargs)
    else:
        perms_modified = False
    return created_dir or perms_modified


class LazyItemsDict(UserDict):
    """A mapping object that behaves like a standard dict except that it allows
    for lazy initialization of values via callable objects.  Lazy items can be
    overwritten and deleted just as normal items."""

    __slots__ = ("lazy_items",)

    def __init__(self, *args, **kwargs):
        self.lazy_items = {}
        UserDict.__init__(self, *args, **kwargs)

    def addLazyItem(self, item_key, value_callable, *pargs, **kwargs):
        """Add a lazy item for the given key.  When the item is requested,
        value_callable will be called with *pargs and **kwargs arguments."""
        self.lazy_items[item_key] = self._LazyItem(value_callable, pargs, kwargs, False)
        # make it show up in self.keys(), etc...
        UserDict.__setitem__(self, item_key, None)

    def addLazySingleton(self, item_key, value_callable, *pargs, **kwargs):
        """This is like addLazyItem except value_callable will only be called
        a maximum of 1 time and the result will be cached for future requests."""
        self.lazy_items[item_key] = self._LazyItem(value_callable, pargs, kwargs, True)
        # make it show up in self.keys(), etc...
        UserDict.__setitem__(self, item_key, None)

    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError(
                "expected at most 1 positional argument, got " + repr(len(args))
            )
        if args:
            map_obj = args[0]
        else:
            map_obj = None
        if map_obj is None:
            pass
        elif isinstance(map_obj, LazyItemsDict):
            for k in map_obj:
                if k in map_obj.lazy_items:
                    UserDict.__setitem__(self, k, None)
                else:
                    UserDict.__setitem__(self, k, map_obj[k])
            self.lazy_items.update(map_obj.lazy_items)
        else:
            UserDict.update(self, map_obj)
        if kwargs:
            UserDict.update(self, kwargs)

    def __getitem__(self, item_key):
        if item_key in self.lazy_items:
            lazy_item = self.lazy_items[item_key]
            pargs = lazy_item.pargs
            if pargs is None:
                pargs = ()
            kwargs = lazy_item.kwargs
            if kwargs is None:
                kwargs = {}
            result = lazy_item.func(*pargs, **kwargs)
            if lazy_item.singleton:
                self[item_key] = result
            return result

        return UserDict.__getitem__(self, item_key)

    def __setitem__(self, item_key, value):
        if item_key in self.lazy_items:
            del self.lazy_items[item_key]
        UserDict.__setitem__(self, item_key, value)

    def __delitem__(self, item_key):
        if item_key in self.lazy_items:
            del self.lazy_items[item_key]
        UserDict.__delitem__(self, item_key)

    def clear(self):
        self.lazy_items.clear()
        UserDict.clear(self)

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return self.__class__(self)

    def __deepcopy__(self, memo=None):
        """
        This forces evaluation of each contained lazy item, and deepcopy of
        the result. A TypeError is raised if any contained lazy item is not
        a singleton, since it is not necessarily possible for the behavior
        of this type of item to be safely preserved.
        """
        if memo is None:
            memo = {}
        result = self.__class__()
        memo[id(self)] = result
        for k in self:
            k_copy = deepcopy(k, memo)
            lazy_item = self.lazy_items.get(k)
            if lazy_item is not None:
                if not lazy_item.singleton:
                    raise TypeError(
                        "LazyItemsDict "
                        + "deepcopy is unsafe with lazy items that are "
                        + f"not singletons: key={k} value={lazy_item}"
                    )
            UserDict.__setitem__(result, k_copy, deepcopy(self[k], memo))
        return result

    class _LazyItem:
        __slots__ = ("func", "pargs", "kwargs", "singleton")

        def __init__(self, func, pargs, kwargs, singleton):
            if not pargs:
                pargs = None
            if not kwargs:
                kwargs = None

            self.func = func
            self.pargs = pargs
            self.kwargs = kwargs
            self.singleton = singleton

        def __copy__(self):
            return self.__class__(self.func, self.pargs, self.kwargs, self.singleton)

        def __deepcopy__(self, memo=None):
            """
            Override this since the default implementation can fail silently,
            leaving some attributes unset.
            """
            if memo is None:
                memo = {}
            result = self.__copy__()
            memo[id(self)] = result
            result.func = deepcopy(self.func, memo)
            result.pargs = deepcopy(self.pargs, memo)
            result.kwargs = deepcopy(self.kwargs, memo)
            result.singleton = deepcopy(self.singleton, memo)
            return result


class ConfigProtect:
    def __init__(self, myroot, protect_list, mask_list, case_insensitive=False):
        self.myroot = myroot
        self.protect_list = protect_list
        self.mask_list = mask_list
        self.case_insensitive = case_insensitive
        self.updateprotect()

    def updateprotect(self):
        """Update internal state for isprotected() calls.  Nonexistent paths
        are ignored."""

        os = _os_merge

        self.protect = []
        self._dirs = set()
        for x in self.protect_list:
            ppath = normalize_path(os.path.join(self.myroot, x.lstrip(os.path.sep)))
            # Protect files that don't exist (bug #523684). If the
            # parent directory doesn't exist, we can safely skip it.
            if os.path.isdir(os.path.dirname(ppath)):
                self.protect.append(ppath)
            try:
                if stat.S_ISDIR(os.stat(ppath).st_mode):
                    self._dirs.add(ppath)
            except OSError:
                pass

        self.protectmask = []
        for x in self.mask_list:
            ppath = normalize_path(os.path.join(self.myroot, x.lstrip(os.path.sep)))
            if self.case_insensitive:
                ppath = ppath.lower()
            try:
                """Use lstat so that anything, even a broken symlink can be
                protected."""
                if stat.S_ISDIR(os.lstat(ppath).st_mode):
                    self._dirs.add(ppath)
                self.protectmask.append(ppath)
                """Now use stat in case this is a symlink to a directory."""
                if stat.S_ISDIR(os.stat(ppath).st_mode):
                    self._dirs.add(ppath)
            except OSError:
                # If it doesn't exist, there's no need to mask it.
                pass

    def isprotected(self, obj):
        """Returns True if obj is protected, False otherwise.  The caller must
        ensure that obj is normalized with a single leading slash.  A trailing
        slash is optional for directories."""
        masked = 0
        protected = 0
        sep = os.path.sep
        if self.case_insensitive:
            obj = obj.lower()
        for ppath in self.protect:
            if len(ppath) > masked and obj.startswith(ppath):
                if ppath in self._dirs:
                    if obj != ppath and not obj.startswith(ppath + sep):
                        # /etc/foo does not match /etc/foobaz
                        continue
                elif obj != ppath:
                    # force exact match when CONFIG_PROTECT lists a
                    # non-directory
                    continue
                protected = len(ppath)
                # config file management
                for pmpath in self.protectmask:
                    if len(pmpath) >= protected and obj.startswith(pmpath):
                        if pmpath in self._dirs:
                            if obj != pmpath and not obj.startswith(pmpath + sep):
                                # /etc/foo does not match /etc/foobaz
                                continue
                        elif obj != pmpath:
                            # force exact match when CONFIG_PROTECT_MASK lists
                            # a non-directory
                            continue
                        # skip, it's in the mask
                        masked = len(pmpath)
        return protected > masked


def new_protect_filename(mydest, newmd5=None, force=False):
    """Resolves a config-protect filename for merging, optionally
    using the last filename if the md5 matches. If force is True,
    then a new filename will be generated even if mydest does not
    exist yet.
    (dest,md5) ==> 'string'            --- path_to_target_filename
    (dest)     ==> ('next', 'highest') --- next_target and most-recent_target
    """

    # config protection filename format:
    # ._cfg0000_foo
    # 0123456789012

    os = _os_merge

    prot_num = -1
    last_pfile = ""

    if not force and not os.path.exists(mydest):
        return mydest

    real_filename = os.path.basename(mydest)
    real_dirname = os.path.dirname(mydest)
    for pfile in os.listdir(real_dirname):
        if pfile[0:5] != "._cfg":
            continue
        if pfile[10:] != real_filename:
            continue
        try:
            new_prot_num = int(pfile[5:9])
            if new_prot_num > prot_num:
                prot_num = new_prot_num
                last_pfile = pfile
        except ValueError:
            continue
    prot_num = prot_num + 1

    new_pfile = normalize_path(
        os.path.join(
            real_dirname, "._cfg" + str(prot_num).zfill(4) + "_" + real_filename
        )
    )
    old_pfile = normalize_path(os.path.join(real_dirname, last_pfile))
    if last_pfile and newmd5:
        try:
            old_pfile_st = os.lstat(old_pfile)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        else:
            if stat.S_ISLNK(old_pfile_st.st_mode):
                try:
                    # Read symlink target as bytes, in case the
                    # target path has a bad encoding.
                    pfile_link = os.readlink(
                        _unicode_encode(
                            old_pfile, encoding=_encodings["merge"], errors="strict"
                        )
                    )
                except OSError:
                    if e.errno != errno.ENOENT:
                        raise
                else:
                    pfile_link = _unicode_decode(
                        pfile_link, encoding=_encodings["merge"], errors="replace"
                    )
                    if pfile_link == newmd5:
                        return old_pfile
            else:
                try:
                    last_pfile_md5 = portage.checksum._perform_md5_merge(old_pfile)
                except FileNotFound:
                    # The file suddenly disappeared or it's a
                    # broken symlink.
                    pass
                else:
                    if last_pfile_md5 == newmd5:
                        return old_pfile
    return new_pfile


def find_updated_config_files(target_root, config_protect):
    """
    Return a tuple of configuration files that needs to be updated.
    The tuple contains lists organized like this:
            [protected_dir, file_list]
    If the protected config isn't a protected_dir but a procted_file, list is:
            [protected_file, None]
    If no configuration files needs to be updated, None is returned
    """

    encoding = _encodings["fs"]

    if config_protect:
        # directories with some protect files in them
        for x in config_protect:
            files = []

            x = os.path.join(target_root, x.lstrip(os.path.sep))
            if not os.access(x, os.W_OK):
                continue
            try:
                mymode = os.lstat(x).st_mode
            except OSError:
                continue

            if stat.S_ISLNK(mymode):
                # We want to treat it like a directory if it
                # is a symlink to an existing directory.
                try:
                    real_mode = os.stat(x).st_mode
                    if stat.S_ISDIR(real_mode):
                        mymode = real_mode
                except OSError:
                    pass

            if stat.S_ISDIR(mymode):
                mycommand = (
                    f"find '{x}' -name '.*' -type d -prune -o -name '._cfg????_*'"
                )
            else:
                mycommand = (
                    "find '%s' -maxdepth 1 -name '._cfg????_%s'"
                    % os.path.split(x.rstrip(os.path.sep))
                )
            mycommand += " ! -name '.*~' ! -iname '.*.bak' -print0"
            cmd = shlex_split(mycommand)

            cmd = [
                _unicode_encode(arg, encoding=encoding, errors="strict") for arg in cmd
            ]
            proc = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            output = _unicode_decode(proc.communicate()[0], encoding=encoding)
            status = proc.wait()
            if os.WIFEXITED(status) and os.WEXITSTATUS(status) == os.EX_OK:
                files = output.split("\0")
                # split always produces an empty string as the last element
                if files and not files[-1]:
                    del files[-1]
                if files:
                    if stat.S_ISDIR(mymode):
                        yield (x, files)
                    else:
                        yield (x, None)


_ld_so_include_re = re.compile(r"^include\s+(\S.*)")


def getlibpaths(root, env=None):
    def read_ld_so_conf(path):
        for l in grabfile(path):
            include_match = _ld_so_include_re.match(l)
            if include_match is not None:
                subpath = os.path.join(os.path.dirname(path), include_match.group(1))
                for p in glob.glob(subpath):
                    yield from read_ld_so_conf(p)
            else:
                yield l

    """ Return a list of paths that are used for library lookups """
    if env is None:
        env = os.environ
    # the following is based on the information from ld.so(8)
    rval = env.get("LD_LIBRARY_PATH", "").split(":")
    rval.extend(read_ld_so_conf(os.path.join(root, "etc", "ld.so.conf")))
    rval.append("/usr/lib")
    rval.append("/lib")

    return [normalize_path(x) for x in rval if x]


def no_color(settings: Optional[dict]) -> bool:
    # In several years (2026+), we can cleanup NOCOLOR support, and just support NO_COLOR.
    has_color: str = settings.get("NO_COLOR")
    nocolor: str = settings.get("NOCOLOR", "false").lower()
    if has_color is None:
        return nocolor in ("yes", "true")
    return bool(has_color)



# testConfigFileSet.py -- Portage Unit Testing Functionality
# Copyright 2007 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import tempfile

from portage import os
from portage.tests import TestCase, test_cps
from portage._sets.files import ConfigFileSet


class ConfigFileSetTestCase(TestCase):
    """Simple Test Case for ConfigFileSet"""

    def setUp(self):
        fd, self.testfile = tempfile.mkstemp(
            suffix=".testdata", prefix=self.__class__.__name__, text=True
        )
        f = os.fdopen(fd, "w")
        for i in range(0, len(test_cps)):
            atom = test_cps[i]
            if i % 2 == 0:
                f.write(atom + " abc def\n")
            else:
                f.write(atom + "\n")
        f.close()

    def tearDown(self):
        os.unlink(self.testfile)

    def testConfigStaticFileSet(self):
        s = ConfigFileSet(self.testfile)
        s.load()
        self.assertEqual(set(test_cps), s.getAtoms())



# Copyright 1999-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.AsynchronousTask import AsynchronousTask
from portage import os


class CompositeTask(AsynchronousTask):
    __slots__ = ("_current_task",)

    _TASK_QUEUED = -1

    def _cancel(self):
        if self._current_task is not None:
            if self._current_task is self._TASK_QUEUED:
                self.returncode = 1
                self._current_task = None
                self._async_wait()
            else:
                self._current_task.cancel()
        elif self.returncode is None:
            # Assume that the task has not started yet.
            self._was_cancelled()
            self._async_wait()

    def _poll(self):
        """
        This does a loop calling self._current_task.poll()
        repeatedly as long as the value of self._current_task
        keeps changing. It calls poll() a maximum of one time
        for a given self._current_task instance. This is useful
        since calling poll() on a task can trigger advance to
        the next task could eventually lead to the returncode
        being set in cases when polling only a single task would
        not have the same effect.
        """

        prev = None
        while True:
            task = self._current_task
            if task is None or task is self._TASK_QUEUED or task is prev:
                # don't poll the same task more than once
                break
            task.poll()
            prev = task

        return self.returncode

    def _assert_current(self, task):
        """
        Raises an AssertionError if the given task is not the
        same one as self._current_task. This can be useful
        for detecting bugs.
        """
        if task is not self._current_task:
            raise AssertionError(f"Unrecognized task: {task}")

    def _default_exit(self, task):
        """
        Calls _assert_current() on the given task and then sets the
        composite returncode attribute if task.returncode != os.EX_OK.
        If the task failed then self._current_task will be set to None.
        Subclasses can use this as a generic task exit callback.

        @rtype: int
        @return: The task.returncode attribute.
        """
        self._assert_current(task)
        if task.returncode != os.EX_OK:
            self.returncode = task.returncode
            self.cancelled = task.cancelled
            self._current_task = None
        return task.returncode

    def _final_exit(self, task):
        """
        Assumes that task is the final task of this composite task.
        Calls _default_exit() and sets self.returncode to the task's
        returncode and sets self._current_task to None.
        """
        self._default_exit(task)
        self._current_task = None
        self.returncode = task.returncode
        return self.returncode

    def _default_final_exit(self, task):
        """
        This calls _final_exit() and then wait().

        Subclasses can use this as a generic final task exit callback.

        """
        self._final_exit(task)
        return self.wait()

    def _start_task(self, task, exit_handler):
        """
        Register exit handler for the given task, set it
        as self._current_task, and call task.start().

        Subclasses can use this as a generic way to start
        a task.

        """
        try:
            task.scheduler = self.scheduler
        except AttributeError:
            pass
        task.addExitListener(exit_handler)
        self._current_task = task
        task.start()

    def _task_queued(self, task):
        task.addStartListener(self._task_queued_start_handler)
        self._current_task = self._TASK_QUEUED

    def _task_queued_start_handler(self, task):
        self._current_task = task

    def _task_queued_wait(self):
        return (
            self._current_task is not self._TASK_QUEUED
            or self.cancelled
            or self.returncode is not None
        )



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools

from _emerge.AsynchronousLock import AsynchronousLock
from _emerge.CompositeTask import CompositeTask
from _emerge.SpawnProcess import SpawnProcess
from urllib.parse import urlparse as urllib_parse_urlparse
import stat
import sys
import portage
from portage import os
from portage.binpkg import get_binpkg_format
from portage.exception import FileNotFound
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from portage.util._pty import _create_pty_or_pipe


class BinpkgFetcher(CompositeTask):
    __slots__ = ("pkg", "pretend", "logfile", "pkg_path", "pkg_allocated_path")

    def __init__(self, **kwargs):
        CompositeTask.__init__(self, **kwargs)

        pkg = self.pkg
        bintree = pkg.root_config.trees["bintree"]
        instance_key = bintree.dbapi._instance_key(pkg.cpv)

        binpkg_path = bintree._remotepkgs[instance_key].get("PATH")
        if not binpkg_path:
            raise FileNotFound(
                f"PATH not found in the binpkg index, the binhost's portage is probably out of date."
            )
        binpkg_format = get_binpkg_format(binpkg_path)

        self.pkg_allocated_path = pkg.root_config.trees["bintree"].getname(
            pkg.cpv, allocate_new=True, remote_binpkg_format=binpkg_format
        )
        self.pkg_path = self.pkg_allocated_path + ".partial"

    def _start(self):
        fetcher = _BinpkgFetcherProcess(
            background=self.background,
            logfile=self.logfile,
            pkg=self.pkg,
            pkg_path=self.pkg_path,
            pretend=self.pretend,
            scheduler=self.scheduler,
        )

        if not self.pretend:
            portage.util.ensure_dirs(os.path.dirname(self.pkg_path))
            if "distlocks" in self.pkg.root_config.settings.features:
                self._start_task(
                    AsyncTaskFuture(future=fetcher.async_lock()),
                    functools.partial(self._start_locked, fetcher),
                )
                return

        self._start_task(fetcher, self._fetcher_exit)

    def _start_locked(self, fetcher, lock_task):
        self._assert_current(lock_task)
        if lock_task.cancelled:
            self._default_final_exit(lock_task)
            return

        lock_task.future.result()
        self._start_task(fetcher, self._fetcher_exit)

    def _fetcher_exit(self, fetcher):
        self._assert_current(fetcher)
        if not self.pretend and fetcher.returncode == os.EX_OK:
            fetcher.sync_timestamp()
        if fetcher.locked:
            self._start_task(
                AsyncTaskFuture(future=fetcher.async_unlock()),
                functools.partial(self._fetcher_exit_unlocked, fetcher),
            )
        else:
            self._fetcher_exit_unlocked(fetcher)

    def _fetcher_exit_unlocked(self, fetcher, unlock_task=None):
        if unlock_task is not None:
            self._assert_current(unlock_task)
            if unlock_task.cancelled:
                self._default_final_exit(unlock_task)
                return

            unlock_task.future.result()

        self._current_task = None
        self.returncode = fetcher.returncode
        self._async_wait()


class _BinpkgFetcherProcess(SpawnProcess):
    __slots__ = ("pkg", "pretend", "locked", "pkg_path", "_lock_obj")

    def _start(self):
        pkg = self.pkg
        pretend = self.pretend
        bintree = pkg.root_config.trees["bintree"]
        settings = bintree.settings
        pkg_path = self.pkg_path

        exists = os.path.exists(pkg_path)
        resume = exists and os.path.basename(pkg_path) in bintree.invalids
        if not (pretend or resume):
            # Remove existing file or broken symlink.
            try:
                os.unlink(pkg_path)
            except OSError:
                pass

        # urljoin doesn't work correctly with
        # unrecognized protocols like sftp
        fetchcommand = None
        resumecommand = None
        if bintree._remote_has_index:
            remote_metadata = bintree._remotepkgs[bintree.dbapi._instance_key(pkg.cpv)]
            rel_uri = remote_metadata.get("PATH")
            if not rel_uri:
                # Assume that the remote index is out of date. No path should
                # never happen in new portage versions.
                rel_uri = pkg.cpv + ".tbz2"
            remote_base_uri = remote_metadata["BASE_URI"]
            uri = remote_base_uri.rstrip("/") + "/" + rel_uri.lstrip("/")
            fetchcommand = remote_metadata.get("FETCHCOMMAND")
            resumecommand = remote_metadata.get("RESUMECOMMAND")
        else:
            raise FileNotFound("Binary packages index not found")

        if pretend:
            portage.writemsg_stdout(f"\n{uri}\n", noiselevel=-1)
            self.returncode = os.EX_OK
            self._async_wait()
            return

        fcmd = None
        if resume:
            fcmd = resumecommand
        else:
            fcmd = fetchcommand
        if fcmd is None:
            protocol = urllib_parse_urlparse(uri)[0]
            fcmd_prefix = "FETCHCOMMAND"
            if resume:
                fcmd_prefix = "RESUMECOMMAND"
            fcmd = settings.get(fcmd_prefix + "_" + protocol.upper())
            if not fcmd:
                fcmd = settings.get(fcmd_prefix)

        fcmd_vars = {
            "DISTDIR": os.path.dirname(pkg_path),
            "URI": uri,
            "FILE": os.path.basename(pkg_path),
        }

        for k in ("PORTAGE_SSH_OPTS",):
            v = settings.get(k)
            if v is not None:
                fcmd_vars[k] = v

        fetch_env = dict(settings.items())
        fetch_args = [
            portage.util.varexpand(x, mydict=fcmd_vars)
            for x in portage.util.shlex_split(fcmd)
        ]

        if self.fd_pipes is None:
            self.fd_pipes = {}
        fd_pipes = self.fd_pipes

        # Redirect all output to stdout since some fetchers like
        # wget pollute stderr (if portage detects a problem then it
        # can send it's own message to stderr).
        fd_pipes.setdefault(0, portage._get_stdin().fileno())
        fd_pipes.setdefault(1, sys.__stdout__.fileno())
        fd_pipes.setdefault(2, sys.__stdout__.fileno())

        self.args = fetch_args
        self.env = fetch_env
        if settings.selinux_enabled():
            self._selinux_type = settings["PORTAGE_FETCH_T"]
        self.log_filter_file = settings.get("PORTAGE_LOG_FILTER_FILE_CMD")
        SpawnProcess._start(self)

    def _pipe(self, fd_pipes):
        """When appropriate, use a pty so that fetcher progress bars,
        like wget has, will work properly."""
        if self.background or not sys.__stdout__.isatty():
            # When the output only goes to a log file,
            # there's no point in creating a pty.
            return os.pipe()
        stdout_pipe = None
        if not self.background:
            stdout_pipe = fd_pipes.get(1)
        got_pty, master_fd, slave_fd = _create_pty_or_pipe(copy_term_size=stdout_pipe)
        return (master_fd, slave_fd)

    def sync_timestamp(self):
        # If possible, update the mtime to match the remote package if
        # the fetcher didn't already do it automatically.
        bintree = self.pkg.root_config.trees["bintree"]
        if bintree._remote_has_index:
            remote_mtime = bintree._remotepkgs[
                bintree.dbapi._instance_key(self.pkg.cpv)
            ].get("_mtime_")
            if remote_mtime is not None:
                try:
                    remote_mtime = int(remote_mtime)
                except ValueError:
                    pass
                else:
                    try:
                        local_mtime = os.stat(self.pkg_path)[stat.ST_MTIME]
                    except OSError:
                        pass
                    else:
                        if remote_mtime != local_mtime:
                            try:
                                os.utime(self.pkg_path, (remote_mtime, remote_mtime))
                            except OSError:
                                pass

    def async_lock(self):
        """
        This raises an AlreadyLocked exception if lock() is called
        while a lock is already held. In order to avoid this, call
        unlock() or check whether the "locked" attribute is True
        or False before calling lock().
        """
        if self._lock_obj is not None:
            raise self.AlreadyLocked((self._lock_obj,))

        result = self.scheduler.create_future()

        def acquired_lock(async_lock):
            if async_lock.wait() == os.EX_OK:
                self.locked = True
                result.set_result(None)
            else:
                result.set_exception(
                    AssertionError(
                        f"AsynchronousLock failed with returncode {async_lock.returncode}"
                    )
                )

        self._lock_obj = AsynchronousLock(path=self.pkg_path, scheduler=self.scheduler)
        self._lock_obj.addExitListener(acquired_lock)
        self._lock_obj.start()
        return result

    class AlreadyLocked(portage.exception.PortageException):
        pass

    def async_unlock(self):
        if self._lock_obj is None:
            raise AssertionError("already unlocked")
        result = self._lock_obj.async_unlock()
        self._lock_obj = None
        self.locked = False
        return result



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop


class FutureDoneCallbackTestCase(TestCase):
    def testFutureDoneCallback(self):
        event_loop = global_event_loop()

        def done_callback(finished):
            done_callback_called.set_result(True)

        done_callback_called = event_loop.create_future()
        finished = event_loop.create_future()
        finished.add_done_callback(done_callback)
        event_loop.call_soon(finished.set_result, True)
        event_loop.run_until_complete(done_callback_called)

        def done_callback2(finished):
            done_callback2_called.set_result(True)

        done_callback_called = event_loop.create_future()
        done_callback2_called = event_loop.create_future()
        finished = event_loop.create_future()
        finished.add_done_callback(done_callback)
        finished.add_done_callback(done_callback2)
        finished.remove_done_callback(done_callback)
        event_loop.call_soon(finished.set_result, True)
        event_loop.run_until_complete(done_callback2_called)

        self.assertFalse(done_callback_called.done())



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import signal
import sys

from portage import _unicode_decode
from portage.output import bold, create_color_func


class UserQuery:
    """The UserQuery class is used to prompt the user with a set of responses,
    as well as accepting and handling the responses."""

    def __init__(self, myopts):
        self.myopts = myopts

    def query(self, prompt, enter_invalid, responses=None, colours=None):
        """Display a prompt and a set of responses, then waits for user input
        and check it against the responses. The first match is returned.

        An empty response will match the first value in the list of responses,
        unless enter_invalid is True. The input buffer is *not* cleared prior
        to the prompt!

        prompt: The String to display as a prompt.
        responses: a List of Strings with the acceptable responses.
        colours: a List of Functions taking and returning a String, used to
        process the responses for display. Typically these will be functions
        like red() but could be e.g. lambda x: "DisplayString".

        If responses is omitted, it defaults to ["Yes", "No"], [green, red].
        If only colours is omitted, it defaults to [bold, ...].

        Returns a member of the List responses. (If called without optional
        arguments, it returns "Yes" or "No".)

        KeyboardInterrupt is converted to SystemExit to avoid tracebacks being
        printed."""
        if responses is None:
            responses = ["Yes", "No"]
            colours = [
                create_color_func("PROMPT_CHOICE_DEFAULT"),
                create_color_func("PROMPT_CHOICE_OTHER"),
            ]
        elif colours is None:
            colours = [bold]
        colours = (colours * len(responses))[: len(responses)]
        responses = [_unicode_decode(x) for x in responses]
        if "--alert" in self.myopts:
            prompt = "\a" + prompt
        print(bold(prompt), end=" ")
        try:
            while True:
                try:
                    response = input(
                        f"[{'/'.join([colours[i](responses[i]) for i in range(len(responses))])}] "
                    )
                except UnicodeDecodeError as e:
                    response = _unicode_decode(e.object).rstrip("\n")
                if response or not enter_invalid:
                    for key in responses:
                        # An empty response will match the
                        # first value in responses.
                        if response.upper() == key[: len(response)].upper():
                            return key
                print(f"Sorry, response '{response}' not understood.", end=" ")
        except (EOFError, KeyboardInterrupt):
            print("Interrupted.")
            sys.exit(128 + signal.SIGINT)



# Copyright 1999-2009 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.BinpkgFetcher import BinpkgFetcher
from _emerge.CompositeTask import CompositeTask
from _emerge.BinpkgVerifier import BinpkgVerifier
from portage import os


class BinpkgPrefetcher(CompositeTask):
    __slots__ = ("pkg",) + (
        "pkg_path",
        "pkg_allocated_path",
        "_bintree",
    )

    def _start(self):
        self._bintree = self.pkg.root_config.trees["bintree"]
        fetcher = BinpkgFetcher(
            background=self.background,
            logfile=self.scheduler.fetch.log_file,
            pkg=self.pkg,
            scheduler=self.scheduler,
        )
        self.pkg_path = fetcher.pkg_path
        self.pkg_allocated_path = fetcher.pkg_allocated_path
        self._start_task(fetcher, self._fetcher_exit)

    def _fetcher_exit(self, fetcher):
        if self._default_exit(fetcher) != os.EX_OK:
            self.wait()
            return

        verifier = BinpkgVerifier(
            background=self.background,
            logfile=self.scheduler.fetch.log_file,
            pkg=self.pkg,
            scheduler=self.scheduler,
            _pkg_path=self.pkg_path,
        )
        self._start_task(verifier, self._verifier_exit)

    def _verifier_exit(self, verifier):
        if self._default_exit(verifier) != os.EX_OK:
            self.wait()
            return

        self._bintree.inject(
            self.pkg.cpv,
            current_pkg_path=self.pkg_path,
            allocated_pkg_path=self.pkg_allocated_path,
        )

        self._current_task = None
        self.returncode = os.EX_OK
        self.wait()



# Copyright 1999-2009 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.DepPriority import DepPriority


class BlockerDepPriority(DepPriority):
    __slots__ = ()

    def __int__(self):
        return 0

    def __str__(self):
        return "blocker"


BlockerDepPriority.instance = BlockerDepPriority()



# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import platform

import pytest

import portage.process
from portage.const import BASH_BINARY


@pytest.mark.skipif(platform.system() != "Linux", reason="not Linux")
def test_spawnE2big(capsys, tmp_path):
    env = dict()
    env["VERY_LARGE_ENV_VAR"] = "X" * 1024 * 256

    logfile = tmp_path / "logfile"
    echo_output = "Should never appear"
    with capsys.disabled():
        retval = portage.process.spawn(
            [BASH_BINARY, "-c", "echo", echo_output], env=env, logfile=logfile
        )

    with open(logfile) as f:
        logfile_content = f.read()
        assert (
            "Largest environment variable: VERY_LARGE_ENV_VAR (262164 bytes)"
            in logfile_content
        )
    assert retval == 1



# elog/mod_custom.py - elog dispatch module
# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import portage.elog.mod_save
import portage.exception
import portage.process


def process(mysettings, key, logentries, fulltext):
    elogfilename = portage.elog.mod_save.process(mysettings, key, logentries, fulltext)

    if not mysettings.get("PORTAGE_ELOG_COMMAND"):
        raise portage.exception.MissingParameter(
            "!!! Custom logging requested but PORTAGE_ELOG_COMMAND is not defined"
        )
    else:
        mylogcmd = mysettings["PORTAGE_ELOG_COMMAND"]
        mylogcmd = mylogcmd.replace("${LOGFILE}", elogfilename)
        mylogcmd = mylogcmd.replace("${PACKAGE}", key)
        retval = portage.process.spawn_bash(mylogcmd)
        if retval != 0:
            raise portage.exception.PortageException(
                "!!! PORTAGE_ELOG_COMMAND failed with exitcode %d" % retval
            )



# Copright Gentoo Foundation 2006-2020
# Portage Unit Testing Functionality

import tempfile
import io
import tarfile
from os import urandom

import portage.gpkg
from portage import os
from portage import shutil
from portage.tests import TestCase
from portage.exception import CompressorOperationFailed


class test_gpkg_stream_case(TestCase):
    def test_gpkg_stream_reader(self):
        data = urandom(1048576)
        data_io = io.BytesIO(data)
        data_io.seek(0)
        with portage.gpkg.tar_stream_reader(data_io, ["cat"]) as test_reader:
            data2 = test_reader.read()
        data_io.close()
        self.assertEqual(data, data2)

    def test_gpkg_stream_reader_without_cmd(self):
        data = urandom(1048576)
        data_io = io.BytesIO(data)
        data_io.seek(0)
        with portage.gpkg.tar_stream_reader(data_io) as test_reader:
            data2 = test_reader.read()
        data_io.close()
        self.assertEqual(data, data2)

    def test_gpkg_stream_reader_kill(self):
        data = urandom(1048576)
        data_io = io.BytesIO(data)
        data_io.seek(0)
        with portage.gpkg.tar_stream_reader(data_io, ["cat"]) as test_reader:
            try:
                test_reader.kill()
            except CompressorOperationFailed:
                pass
        data_io.close()
        self.assertNotEqual(test_reader.proc.poll(), None)

    def test_gpkg_stream_reader_kill_without_cmd(self):
        data = urandom(1048576)
        data_io = io.BytesIO(data)
        data_io.seek(0)
        with portage.gpkg.tar_stream_reader(data_io) as test_reader:
            test_reader.kill()
        data_io.close()
        self.assertEqual(test_reader.proc, None)

    def test_gpkg_stream_writer(self):
        tmpdir = tempfile.mkdtemp()
        try:
            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            data = urandom(1048576)
            with tarfile.open(gpkg_file_loc, "w") as test_tar:
                test_tarinfo = tarfile.TarInfo("test")
                with portage.gpkg.tar_stream_writer(
                    test_tarinfo, test_tar, tarfile.USTAR_FORMAT, ["cat"]
                ) as test_writer:
                    test_writer.write(data)

            with tarfile.open(gpkg_file_loc, "r") as test_tar:
                test_tarinfo = test_tar.getmember("test")
                data2 = test_tar.extractfile(test_tarinfo).read()
                self.assertEqual(data, data2)
        finally:
            shutil.rmtree(tmpdir)

    def test_gpkg_stream_writer_without_cmd(self):
        tmpdir = tempfile.mkdtemp()

        try:
            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            data = urandom(1048576)
            with tarfile.open(gpkg_file_loc, "w") as test_tar:
                test_tarinfo = tarfile.TarInfo("test")
                with portage.gpkg.tar_stream_writer(
                    test_tarinfo, test_tar, tarfile.USTAR_FORMAT
                ) as test_writer:
                    test_writer.write(data)

            with tarfile.open(gpkg_file_loc, "r") as test_tar:
                test_tarinfo = test_tar.getmember("test")
                data2 = test_tar.extractfile(test_tarinfo).read()
                self.assertEqual(data, data2)
        finally:
            shutil.rmtree(tmpdir)



# Copyright 2021 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import _shell_quote
from portage.tests import TestCase


class ShellQuoteTestCase(TestCase):
    def testShellQuote(self):
        test_data = [
            # String contains no special characters, should be preserved.
            ("abc", "abc"),
            # String contains whitespace, should be double-quoted to prevent word splitting.
            ("abc xyz", '"abc xyz"'),
            ("abc  xyz", '"abc  xyz"'),
            (" abcxyz ", '" abcxyz "'),
            ("abc\txyz", '"abc\txyz"'),
            ("abc\t\txyz", '"abc\t\txyz"'),
            ("\tabcxyz\t", '"\tabcxyz\t"'),
            ("abc\nxyz", '"abc\nxyz"'),
            ("abc\n\nxyz", '"abc\n\nxyz"'),
            ("\nabcxyz\n", '"\nabcxyz\n"'),
            # String contains > or <, should be double-quoted to prevent redirection.
            ("abc>xyz", '"abc>xyz"'),
            ("abc>>xyz", '"abc>>xyz"'),
            (">abcxyz>", '">abcxyz>"'),
            ("abc<xyz", '"abc<xyz"'),
            ("abc<<xyz", '"abc<<xyz"'),
            ("<abcxyz<", '"<abcxyz<"'),
            # String contains =, should be double-quoted to prevent assignment.
            ("abc=xyz", '"abc=xyz"'),
            ("abc==xyz", '"abc==xyz"'),
            ("=abcxyz=", '"=abcxyz="'),
            # String contains *, should be double-quoted to prevent globbing.
            ("abc*xyz", '"abc*xyz"'),
            ("abc**xyz", '"abc**xyz"'),
            ("*abcxyz*", '"*abcxyz*"'),
            # String contains $, should be escaped to prevent parameter expansion.
            # Also double-quoted, though not strictly needed.
            ("abc$xyz", '"abc\\$xyz"'),
            ("abc$$xyz", '"abc\\$\\$xyz"'),
            ("$abcxyz$", '"\\$abcxyz\\$"'),
            # String contains `, should be escaped to prevent command substitution.
            # Also double-quoted, though not strictly needed.
            ("abc`xyz", '"abc\\`xyz"'),
            ("abc``xyz", '"abc\\`\\`xyz"'),
            ("`abc`", '"\\`abc\\`"'),
            # String contains \, should be escaped to prevent it from escaping
            # the next character. Also double-quoted, though not strictly needed.
            ("abc\\xyz", '"abc\\\\xyz"'),
            ("abc\\\\xyz", '"abc\\\\\\\\xyz"'),
            ("\\abcxyz\\", '"\\\\abcxyz\\\\"'),
            # String contains ", should be escaped to prevent it from unexpectedly
            # ending a previous double-quote or starting a new double-quote. Also
            # double-quoted, though not strictly needed.
            ('abc"xyz', '"abc\\"xyz"'),
            ('abc""xyz', '"abc\\"\\"xyz"'),
            ('"abcxyz"', '"\\"abcxyz\\""'),
            # String contains ', should be double-quoted to prevent it from unexpectedly
            # ending a previous single-quote or starting a new single-quote.
            ("abc'xyz", '"abc\'xyz"'),
            ("abc''xyz", "\"abc''xyz\""),
            ("'abcxyz'", "\"'abcxyz'\""),
            # String contains ;, should be double-quoted to prevent command separation.
            ("abc;xyz", '"abc;xyz"'),
            ("abc;;xyz", '"abc;;xyz"'),
            (";abcxyz;", '";abcxyz;"'),
            # String contains &, should be double-quoted to prevent job control.
            ("abc&xyz", '"abc&xyz"'),
            ("abc&&xyz", '"abc&&xyz"'),
            ("&abcxyz&", '"&abcxyz&"'),
            # String contains |, should be double-quoted to prevent piping.
            ("abc|xyz", '"abc|xyz"'),
            ("abc||xyz", '"abc||xyz"'),
            ("|abcxyz|", '"|abcxyz|"'),
            # String contains (), should be double-quoted to prevent
            # command group / array initialization.
            ("abc()xyz", '"abc()xyz"'),
            ("abc(())xyz", '"abc(())xyz"'),
            ("((abcxyz))", '"((abcxyz))"'),
            # String contains {}. Parameter expansion of the form ${} is already
            # rendered safe by escaping the $, but {} could also occur on its own,
            # for example in a brace expansion such as filename.{ext1,ext2},
            # so the string should be double-quoted.
            ("abc{}xyz", '"abc{}xyz"'),
            ("abc{{}}xyz", '"abc{{}}xyz"'),
            ("{{abcxyz}}", '"{{abcxyz}}"'),
            # String contains [], should be double-quoted to prevent testing
            ("abc[]xyz", '"abc[]xyz"'),
            ("abc[[]]xyz", '"abc[[]]xyz"'),
            ("[[abcxyz]]", '"[[abcxyz]]"'),
            # String contains #, should be double-quoted to prevent comment.
            ("#abc", '"#abc"'),
            # String contains !, should be double-quoted to prevent e.g. history substitution.
            ("!abc", '"!abc"'),
            # String contains ~, should be double-quoted to prevent home directory expansion.
            ("~abc", '"~abc"'),
            # String contains ?, should be double-quoted to prevent globbing.
            ("abc?xyz", '"abc?xyz"'),
            ("abc??xyz", '"abc??xyz"'),
            ("?abcxyz?", '"?abcxyz?"'),
        ]

        for data, expected_result in test_data:
            result = _shell_quote(data)
            self.assertEqual(result, expected_result)



# Copyright 2010-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

__all__ = (
    "ordered_by_atom_specificity",
    "prune_incremental",
)

from _emerge.Package import Package
from portage.dep import best_match_to_list, _repo_separator


def ordered_by_atom_specificity(cpdict, pkg, repo=None):
    """
    Return a list of matched values from the given cpdict,
    in ascending order by atom specificity. The rationale
    for this order is that package.* config files are
    typically written in ChangeLog like fashion, so it's
    most friendly if the order that the atoms are written
    does not matter. Therefore, settings from more specific
    atoms override those of less specific atoms. Without
    this behavior, settings from relatively unspecific atoms
    would (somewhat confusingly) override the settings of
    more specific atoms, requiring people to make adjustments
    to the order that atoms are listed in the config file in
    order to achieve desired results (and thus corrupting
    the ChangeLog like ordering of the file).
    """
    if not hasattr(pkg, "repo") and repo and repo != Package.UNKNOWN_REPO:
        pkg = pkg + _repo_separator + repo

    results = []
    keys = list(cpdict)

    while keys:
        bestmatch = best_match_to_list(pkg, keys)
        if bestmatch:
            keys.remove(bestmatch)
            results.append(cpdict[bestmatch])
        else:
            break

    if results:
        # reverse, so the most specific atoms come last
        results.reverse()

    return results


def prune_incremental(split):
    """
    Prune off any parts of an incremental variable that are
    made irrelevant by the latest occuring * or -*. This
    could be more aggressive but that might be confusing
    and the point is just to reduce noise a bit.
    """
    for i, x in enumerate(reversed(split)):
        if x == "*":
            split = split[-i - 1 :]
            break
        elif x == "-*":
            if i == 0:
                # Preserve the last -*, since otherwise an empty value
                # would trigger fallback to a default value.
                split = ["-*"]
            else:
                split = split[-i:]
            break
    return split



# Copyright 1999-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import logging
from portage import installation
from portage.util import writemsg_level


def create_depgraph_params(myopts, myaction):
    # configure emerge engine parameters
    #
    # autounmask:               enable autounmask
    # autounmask_keep_keywords: prevent autounmask changes to package.accept_keywords
    # autounmask_keep_license:  prevent autounmask changes to package.license
    # autounmask_keep_masks:    prevent autounmask changes to package.mask
    # autounmask_keep_use:      prevent autounmask changes to package.use
    # self:      include _this_ package regardless of if it is merged.
    # selective: exclude the package if it is merged
    # recurse:   go into the dependencies
    # deep:      go into the dependencies of already merged packages
    # empty:     pretend nothing is merged
    # complete:  completely account for all known dependencies
    # bdeps:     satisfy build time dependencies of packages that are
    #   already built, even though they are not strictly required
    # remove:    build graph for use in removing packages
    # rebuilt_binaries: replace installed packages with rebuilt binaries
    # rebuild_if_new_slot: rebuild or reinstall packages when
    # 	slot/sub-slot := operator dependencies can be satisfied by a newer
    # 	slot/sub-slot, so that older packages slots will become eligible for
    # 	removal by the --depclean action as soon as possible
    # ignore_built_slot_operator_deps: ignore the slot/sub-slot := operator parts
    # 	of dependencies that have been recorded when packages where built
    # ignore_soname_deps: ignore the soname dependencies of built
    #   packages, so that they do not trigger dependency resolution
    #   failures, or cause packages to be rebuilt or replaced.
    # ignore_world: ignore the @world package set and its dependencies
    # implicit_system_deps: Assume that packages may have implicit dependencies
    #   on packages which belong to the @system set.
    # with_test_deps: pull in test deps for packages matched by arguments
    # changed_deps: rebuild installed packages with outdated deps
    # changed_deps_report: report installed packages with outdated deps
    # changed_slot: rebuild installed packages with outdated SLOT metadata
    # binpkg_changed_deps: reject binary packages with outdated deps
    myparams = {"recurse": True}

    binpkg_respect_use = myopts.get("--binpkg-respect-use")
    if binpkg_respect_use is not None:
        myparams["binpkg_respect_use"] = binpkg_respect_use
    elif "--usepkgonly" not in myopts:
        # If --binpkg-respect-use is not explicitly specified, we enable
        # the behavior automatically (like requested in bug #297549), as
        # long as it doesn't strongly conflict with other options that
        # have been specified.
        myparams["binpkg_respect_use"] = "auto"

    autounmask_keep_keywords = myopts.get("--autounmask-keep-keywords")
    autounmask_keep_masks = myopts.get("--autounmask-keep-masks")

    autounmask = myopts.get("--autounmask")
    autounmask_license = myopts.get(
        "--autounmask-license", "y" if autounmask is True else "n"
    )
    autounmask_use = (
        "n"
        if myparams.get("binpkg_respect_use") == "y"
        else myopts.get("--autounmask-use")
    )
    if autounmask == "n":
        autounmask = False
    else:
        if autounmask is None:
            if autounmask_use in (None, "y"):
                autounmask = True
            if autounmask_license in ("y",):
                autounmask = True

            # Do not enable package.accept_keywords or package.mask
            # changes by default.
            if autounmask_keep_keywords is None:
                autounmask_keep_keywords = True
            if autounmask_keep_masks is None:
                autounmask_keep_masks = True
        else:
            autounmask = True

    myparams["autounmask"] = autounmask
    myparams["autounmask_keep_use"] = True if autounmask_use == "n" else False
    myparams["autounmask_keep_license"] = False if autounmask_license == "y" else True
    myparams["autounmask_keep_keywords"] = (
        False if autounmask_keep_keywords in (None, "n") else True
    )
    myparams["autounmask_keep_masks"] = (
        False if autounmask_keep_masks in (None, "n") else True
    )

    bdeps = myopts.get("--with-bdeps")
    if bdeps is not None:
        myparams["bdeps"] = bdeps
    elif myaction == "remove" or (
        myopts.get("--with-bdeps-auto") != "n" and "--usepkg" not in myopts
    ):
        myparams["bdeps"] = "auto"

    ignore_built_slot_operator_deps = myopts.get("--ignore-built-slot-operator-deps")
    if ignore_built_slot_operator_deps is not None:
        myparams["ignore_built_slot_operator_deps"] = ignore_built_slot_operator_deps

    myparams["ignore_soname_deps"] = myopts.get("--ignore-soname-deps", "y")

    dyn_deps_def = (
        "y" if installation.TYPE == installation.TYPES.SOURCE else "@DYN_DEPS_DEFAULT@"
    )
    dynamic_deps = (
        myopts.get("--dynamic-deps", dyn_deps_def) != "n" and "--nodeps" not in myopts
    )
    if dynamic_deps:
        myparams["dynamic_deps"] = True

    myparams["implicit_system_deps"] = myopts.get("--implicit-system-deps", "y") != "n"

    if myaction == "remove":
        myparams["remove"] = True
        myparams["complete"] = True
        myparams["selective"] = True
        return myparams

    if myopts.get("--ignore-world") is True:
        myparams["ignore_world"] = True

    rebuild_if_new_slot = myopts.get("--rebuild-if-new-slot")
    if rebuild_if_new_slot is not None:
        myparams["rebuild_if_new_slot"] = rebuild_if_new_slot

    changed_slot = myopts.get("--changed-slot") is True
    if changed_slot:
        myparams["changed_slot"] = True

    # --update-if-installed implies --update
    update_if_installed = myopts.get("--update-if-installed")
    if update_if_installed is not None:
        myparams["update_if_installed"] = update_if_installed
        myopts["--update"] = True

    if (
        "--update" in myopts
        or "--newrepo" in myopts
        or "--newuse" in myopts
        or "--reinstall" in myopts
        or "--noreplace" in myopts
        or myopts.get("--changed-deps", "n") != "n"
        or changed_slot
        or myopts.get("--selective", "n") != "n"
    ):
        myparams["selective"] = True

    deep = myopts.get("--deep")
    if deep is not None and deep != 0:
        myparams["deep"] = deep

    complete_if_new_use = myopts.get("--complete-graph-if-new-use")
    if complete_if_new_use is not None:
        myparams["complete_if_new_use"] = complete_if_new_use

    complete_if_new_ver = myopts.get("--complete-graph-if-new-ver")
    if complete_if_new_ver is not None:
        myparams["complete_if_new_ver"] = complete_if_new_ver

    if (
        "--complete-graph" in myopts
        or "--rebuild-if-new-rev" in myopts
        or "--rebuild-if-new-ver" in myopts
        or "--rebuild-if-unbuilt" in myopts
    ):
        myparams["complete"] = True
    if "--emptytree" in myopts:
        myparams["empty"] = True
        myparams["deep"] = True
        myparams.pop("selective", None)

    if "--nodeps" in myopts:
        myparams.pop("recurse", None)
        myparams.pop("deep", None)
        myparams.pop("complete", None)

    rebuilt_binaries = myopts.get("--rebuilt-binaries")
    if (
        rebuilt_binaries is True
        or rebuilt_binaries != "n"
        and "--usepkgonly" in myopts
        and myopts.get("--deep") is True
        and "--update" in myopts
    ):
        myparams["rebuilt_binaries"] = True

    binpkg_changed_deps = myopts.get("--binpkg-changed-deps")
    if binpkg_changed_deps is not None:
        myparams["binpkg_changed_deps"] = binpkg_changed_deps
    elif "--usepkgonly" not in myopts:
        # In order to avoid dependency resolution issues due to changed
        # dependencies, enable this automatically, as long as it doesn't
        # strongly conflict with other options that have been specified.
        myparams["binpkg_changed_deps"] = "auto"

    changed_deps = myopts.get("--changed-deps")
    if changed_deps is not None:
        myparams["changed_deps"] = changed_deps

    changed_deps_report = myopts.get("--changed-deps-report", "n") == "y"
    if changed_deps_report:
        myparams["changed_deps_report"] = True

    if myopts.get("--selective") == "n":
        # --selective=n can be used to remove selective
        # behavior that may have been implied by some
        # other option like --update.
        myparams.pop("selective", None)

    with_test_deps = myopts.get("--with-test-deps")
    if with_test_deps is not None:
        myparams["with_test_deps"] = with_test_deps

    if "--debug" in myopts:
        writemsg_level(
            f"\n\nmyparams {myparams}\n\n", noiselevel=-1, level=logging.DEBUG
        )

    return myparams



# Copyright 1999-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import argparse
import locale
import platform
import sys

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "logging",
    "portage.dep:Atom",
    "portage.util:writemsg_level",
    "textwrap",
    "_emerge.actions:load_emerge_config,run_action," + "validate_ebuild_environment",
    "_emerge.help:emerge_help",
    "_emerge.is_valid_package_atom:insert_category_into_atom",
)
from portage import os
from portage.sync import _SUBMODULE_PATH_MAP

from typing import Optional

options = [
    "--alphabetical",
    "--ask-enter-invalid",
    "--buildpkgonly",
    "--changed-use",
    "--columns",
    "--debug",
    "--digest",
    "--emptytree",
    "--verbose-conflicts",
    "--fetchonly",
    "--fetch-all-uri",
    "--ignore-default-opts",
    "--noconfmem",
    "--newrepo",
    "--newuse",
    "--nodeps",
    "--noreplace",
    "--nospinner",
    "--oneshot",
    "--onlydeps",
    "--pretend",
    "--quiet-repo-display",
    "--quiet-unmerge-warn",
    "--resume",
    "--searchdesc",
    "--skipfirst",
    "--tree",
    "--unordered-display",
    "--update",
    "--update-if-installed",
]

shortmapping = {
    "1": "--oneshot",
    "B": "--buildpkgonly",
    "c": "--depclean",
    "C": "--unmerge",
    "d": "--debug",
    "e": "--emptytree",
    "f": "--fetchonly",
    "F": "--fetch-all-uri",
    "h": "--help",
    "n": "--noreplace",
    "N": "--newuse",
    "o": "--onlydeps",
    "O": "--nodeps",
    "p": "--pretend",
    "P": "--prune",
    "r": "--resume",
    "s": "--search",
    "S": "--searchdesc",
    "t": "--tree",
    "u": "--update",
    "U": "--changed-use",
    "V": "--version",
}

COWSAY_MOO = r"""

  Larry loves Gentoo ({})

 _______________________
< Have you mooed today? >
 -----------------------
        \   ^__^
         \  (oo)\_______
            (__)\       )\/\
                ||----w |
                ||     ||

"""


def multiple_actions(action1, action2):
    sys.stderr.write("\n!!! Multiple actions requested... Please choose one only.\n")
    sys.stderr.write(f"!!! '{action1}' or '{action2}'\n\n")
    sys.exit(1)


def insert_optional_args(args):
    """
    Parse optional arguments and insert a value if one has
    not been provided. This is done before feeding the args
    to the optparse parser since that parser does not support
    this feature natively.
    """

    class valid_integers:
        def __contains__(self, s):
            try:
                return int(s) >= 0
            except (ValueError, OverflowError):
                return False

    valid_integers = valid_integers()

    class valid_floats:
        def __contains__(self, s):
            try:
                return float(s) >= 0
            except (ValueError, OverflowError):
                return False

    valid_floats = valid_floats()

    y_or_n = (
        "y",
        "n",
    )

    new_args = []

    default_arg_opts = {
        "--alert": y_or_n,
        "--ask": y_or_n,
        "--autounmask": y_or_n,
        "--autounmask-continue": y_or_n,
        "--autounmask-only": y_or_n,
        "--autounmask-keep-keywords": y_or_n,
        "--autounmask-keep-masks": y_or_n,
        "--autounmask-unrestricted-atoms": y_or_n,
        "--autounmask-write": y_or_n,
        "--binpkg-changed-deps": y_or_n,
        "--buildpkg": y_or_n,
        "--changed-deps": y_or_n,
        "--changed-slot": y_or_n,
        "--changed-deps-report": y_or_n,
        "--complete-graph": y_or_n,
        "--deep": valid_integers,
        "--depclean-lib-check": y_or_n,
        "--deselect": y_or_n,
        "--binpkg-respect-use": y_or_n,
        "--fail-clean": y_or_n,
        "--fuzzy-search": y_or_n,
        "--getbinpkg": y_or_n,
        "--getbinpkgonly": y_or_n,
        "--ignore-world": y_or_n,
        "--jobs": valid_integers,
        "--keep-going": y_or_n,
        "--load-average": valid_floats,
        "--onlydeps-with-ideps": y_or_n,
        "--onlydeps-with-rdeps": y_or_n,
        "--package-moves": y_or_n,
        "--quiet": y_or_n,
        "--quiet-build": y_or_n,
        "--quiet-fail": y_or_n,
        "--read-news": y_or_n,
        "--rebuild-if-new-slot": y_or_n,
        "--rebuild-if-new-rev": y_or_n,
        "--rebuild-if-new-ver": y_or_n,
        "--rebuild-if-unbuilt": y_or_n,
        "--rebuilt-binaries": y_or_n,
        "--root-deps": ("rdeps",),
        "--select": y_or_n,
        "--selective": y_or_n,
        "--use-ebuild-visibility": y_or_n,
        "--usepkg": y_or_n,
        "--usepkgonly": y_or_n,
        "--usepkg-exclude-live": y_or_n,
        "--verbose": y_or_n,
        "--verbose-slot-rebuilds": y_or_n,
        "--with-test-deps": y_or_n,
    }

    short_arg_opts = {
        "D": valid_integers,
        "j": valid_integers,
        "l": valid_floats,
    }

    # Don't make things like "-kn" expand to "-k n"
    # since existence of -n makes it too ambiguous.
    short_arg_opts_n = {
        "a": y_or_n,
        "A": y_or_n,
        "b": y_or_n,
        "g": y_or_n,
        "G": y_or_n,
        "k": y_or_n,
        "K": y_or_n,
        "q": y_or_n,
        "v": y_or_n,
        "w": y_or_n,
        "W": y_or_n,
    }

    arg_stack = args[:]
    arg_stack.reverse()
    while arg_stack:
        arg = arg_stack.pop()

        default_arg_choices = default_arg_opts.get(arg)
        if default_arg_choices is not None:
            new_args.append(arg)
            if arg_stack and arg_stack[-1] in default_arg_choices:
                new_args.append(arg_stack.pop())
            else:
                # insert default argument
                new_args.append("True")
            continue

        if arg[:1] != "-" or arg[:2] == "--":
            new_args.append(arg)
            continue

        match = None
        for k, arg_choices in short_arg_opts.items():
            if k in arg:
                match = k
                break

        if match is None:
            for k, arg_choices in short_arg_opts_n.items():
                if k in arg:
                    match = k
                    break

        if match is None:
            new_args.append(arg)
            continue

        if len(arg) == 2:
            new_args.append(arg)
            if arg_stack and arg_stack[-1] in arg_choices:
                new_args.append(arg_stack.pop())
            else:
                # insert default argument
                new_args.append("True")
            continue

        # Insert an empty placeholder in order to
        # satisfy the requirements of optparse.

        new_args.append("-" + match)
        opt_arg = None
        saved_opts = None

        if arg[1:2] == match:
            if match not in short_arg_opts_n and arg[2:] in arg_choices:
                opt_arg = arg[2:]
            else:
                saved_opts = arg[2:]
                opt_arg = "True"
        else:
            saved_opts = arg[1:].replace(match, "")
            opt_arg = "True"

        if opt_arg is None and arg_stack and arg_stack[-1] in arg_choices:
            opt_arg = arg_stack.pop()

        if opt_arg is None:
            new_args.append("True")
        else:
            new_args.append(opt_arg)

        if saved_opts is not None:
            # Recycle these on arg_stack since they
            # might contain another match.
            arg_stack.append("-" + saved_opts)

    return new_args


def _find_bad_atoms(atoms, less_strict=False):
    """
    Declares all atoms as invalid that have an operator,
    a use dependency, a blocker or a repo spec.
    It accepts atoms with wildcards.
    In less_strict mode it accepts operators and repo specs.
    """
    bad_atoms = []
    for x in " ".join(atoms).split():
        atom = x
        if "/" not in x.split(":")[0]:
            x_cat = insert_category_into_atom(x, "dummy-category")
            if x_cat is not None:
                atom = x_cat

        bad_atom = False
        try:
            atom = Atom(atom, allow_wildcard=True, allow_repo=less_strict)
        except portage.exception.InvalidAtom:
            bad_atom = True

        if bad_atom or (atom.operator and not less_strict) or atom.blocker or atom.use:
            bad_atoms.append(x)
    return bad_atoms


def parse_opts(tmpcmdline, silent=False):
    myaction = None
    myopts = {}

    actions = frozenset(
        [
            "clean",
            "check-news",
            "config",
            "depclean",
            "help",
            "info",
            "list-sets",
            "metadata",
            "moo",
            "prune",
            "rage-clean",
            "regen",
            "search",
            "sync",
            "unmerge",
            "version",
        ]
    )

    longopt_aliases = {"--cols": "--columns", "--skip-first": "--skipfirst"}
    y_or_n = ("y", "n")
    true_y_or_n = ("True", "y", "n")
    true_y = ("True", "y")
    argument_options = {
        "--alert": {
            "shortopt": "-A",
            "help": "alert (terminal bell) on prompts",
            "choices": true_y_or_n,
        },
        "--ask": {
            "shortopt": "-a",
            "help": "prompt before performing any actions",
            "choices": true_y_or_n,
        },
        "--autounmask": {
            "help": "automatically unmask packages",
            "choices": true_y_or_n,
        },
        "--autounmask-backtrack": {
            "help": (
                "continue backtracking when there are autounmask "
                "configuration changes"
            ),
            "choices": ("y", "n"),
        },
        "--autounmask-continue": {
            "help": "write autounmask changes and continue",
            "choices": true_y_or_n,
        },
        "--autounmask-only": {
            "help": "only perform --autounmask",
            "choices": true_y_or_n,
        },
        "--autounmask-license": {
            "help": "allow autounmask to change package.license",
            "choices": y_or_n,
        },
        "--autounmask-unrestricted-atoms": {
            "help": "write autounmask changes with >= atoms if possible",
            "choices": true_y_or_n,
        },
        "--autounmask-use": {
            "help": "allow autounmask to change package.use",
            "choices": y_or_n,
        },
        "--autounmask-keep-keywords": {
            "help": "don't add package.accept_keywords entries",
            "choices": true_y_or_n,
        },
        "--autounmask-keep-masks": {
            "help": "don't add package.unmask entries",
            "choices": true_y_or_n,
        },
        "--autounmask-write": {
            "help": "write changes made by --autounmask to disk",
            "choices": true_y_or_n,
        },
        "--accept-properties": {
            "help": "temporarily override ACCEPT_PROPERTIES",
            "action": "store",
        },
        "--accept-restrict": {
            "help": "temporarily override ACCEPT_RESTRICT",
            "action": "store",
        },
        "--backtrack": {
            "help": "Specifies how many times to backtrack if dependency "
            + "calculation fails ",
            "action": "store",
        },
        "--binpkg-changed-deps": {
            "help": ("reject binary packages with outdated " "dependencies"),
            "choices": true_y_or_n,
        },
        "--buildpkg": {
            "shortopt": "-b",
            "help": "build binary packages",
            "choices": true_y_or_n,
        },
        "--buildpkg-exclude": {
            "help": "A space separated list of package atoms for which "
            + "no binary packages should be built. This option overrides all "
            + "possible ways to enable building of binary packages.",
            "action": "append",
        },
        "--changed-deps": {
            "help": ("replace installed packages with " "outdated dependencies"),
            "choices": true_y_or_n,
        },
        "--changed-deps-report": {
            "help": ("report installed packages with " "outdated dependencies"),
            "choices": true_y_or_n,
        },
        "--changed-slot": {
            "help": ("replace installed packages with " "outdated SLOT metadata"),
            "choices": true_y_or_n,
        },
        "--config-root": {
            "help": "specify the location for portage configuration files",
            "action": "store",
        },
        "--color": {"help": "enable or disable color output", "choices": ("y", "n")},
        "--complete-graph": {
            "help": "completely account for all known dependencies",
            "choices": true_y_or_n,
        },
        "--complete-graph-if-new-use": {
            "help": "trigger --complete-graph behavior if USE or IUSE will change for an installed package",
            "choices": y_or_n,
        },
        "--complete-graph-if-new-ver": {
            "help": "trigger --complete-graph behavior if an installed package version will change (upgrade or downgrade)",
            "choices": y_or_n,
        },
        "--deep": {
            "shortopt": "-D",
            "help": "Specifies how deep to recurse into dependencies "
            + "of packages given as arguments. If no argument is given, "
            + "depth is unlimited. Default behavior is to skip "
            + "dependencies of installed packages.",
            "action": "store",
        },
        "--depclean-lib-check": {
            "help": "check for consumers of libraries before removing them",
            "choices": true_y_or_n,
        },
        "--deselect": {
            "shortopt": "-W",
            "help": "remove atoms/sets from the world file",
            "choices": true_y_or_n,
        },
        "--dynamic-deps": {
            "help": "substitute the dependencies of installed packages with the dependencies of unbuilt ebuilds",
            "choices": y_or_n,
        },
        "--exclude": {
            "shortopt": "-X",
            "help": "A space separated list of package names or slot atoms. "
            + "Emerge won't  install any ebuild or binary package that "
            + "matches any of the given package atoms.",
            "action": "append",
        },
        "--fail-clean": {
            "help": "clean temp files after build failure",
            "choices": true_y_or_n,
        },
        "--fuzzy-search": {
            "help": "Enable or disable fuzzy search",
            "choices": true_y_or_n,
        },
        "--ignore-built-slot-operator-deps": {
            "help": "Ignore the slot/sub-slot := operator parts of dependencies that have "
            "been recorded when packages where built. This option is intended "
            "only for debugging purposes, and it only affects built packages "
            "that specify slot/sub-slot := operator dependencies using the "
            'experimental "4-slot-abi" EAPI.',
            "choices": y_or_n,
        },
        "--ignore-soname-deps": {
            "help": "Ignore the soname dependencies of binary and "
            "installed packages. This option is enabled by "
            "default, since soname dependencies are relatively "
            "new, and the required metadata is not guaranteed to "
            "exist for binary and installed packages built with "
            "older versions of portage.",
            "choices": y_or_n,
        },
        "--ignore-world": {
            "help": "ignore the @world package set and its dependencies",
            "choices": true_y_or_n,
        },
        "--implicit-system-deps": {
            "help": "Assume that packages may have implicit dependencies on"
            "packages which belong to the @system set",
            "choices": y_or_n,
        },
        "--jobs": {
            "shortopt": "-j",
            "help": "Specifies the number of packages to build " + "simultaneously.",
            "action": "store",
        },
        "--keep-going": {
            "help": "continue as much as possible after an error",
            "choices": true_y_or_n,
        },
        "--load-average": {
            "shortopt": "-l",
            "help": "Specifies that no new builds should be started "
            + "if there are other builds running and the load average "
            + "is at least LOAD (a floating-point number).",
            "action": "store",
        },
        "--misspell-suggestions": {
            "help": "enable package name misspell suggestions",
            "choices": ("y", "n"),
        },
        "--with-bdeps": {
            "help": "include unnecessary build time dependencies",
            "choices": ("y", "n"),
        },
        "--with-bdeps-auto": {
            "help": (
                "automatically enable --with-bdeps for installation"
                " actions, unless --usepkg is enabled"
            ),
            "choices": ("y", "n"),
        },
        "--reinstall": {
            "help": "specify conditions to trigger package reinstallation",
            "choices": ["changed-use"],
        },
        "--reinstall-atoms": {
            "help": "A space separated list of package names or slot atoms. "
            + "Emerge will treat matching packages as if they are not "
            + "installed, and reinstall them if necessary. Implies --deep.",
            "action": "append",
        },
        "--binpkg-respect-use": {
            "help": "discard binary packages if their use flags \
				don't match the current configuration",
            "choices": true_y_or_n,
        },
        "--getbinpkg": {
            "shortopt": "-g",
            "help": "fetch binary packages",
            "choices": true_y_or_n,
        },
        "--getbinpkgonly": {
            "shortopt": "-G",
            "help": "fetch binary packages only",
            "choices": true_y_or_n,
        },
        "--usepkg-exclude": {
            "help": "A space separated list of package names or slot atoms. "
            + "Emerge will ignore matching binary packages. ",
            "action": "append",
        },
        "--onlydeps-with-ideps": {
            "help": "modify interpretation of dependencies to include IDEPEND",
            "choices": true_y_or_n,
        },
        "--onlydeps-with-rdeps": {
            "help": "modify interpretation of dependencies",
            "choices": true_y_or_n,
        },
        "--rebuild-exclude": {
            "help": "A space separated list of package names or slot atoms. "
            + "Emerge will not rebuild these packages due to the "
            + "--rebuild flag. ",
            "action": "append",
        },
        "--rebuild-ignore": {
            "help": "A space separated list of package names or slot atoms. "
            + "Emerge will not rebuild packages that depend on matching "
            + "packages due to the --rebuild flag. ",
            "action": "append",
        },
        "--package-moves": {
            "help": "perform package moves when necessary",
            "choices": true_y_or_n,
        },
        "--prefix": {"help": "specify the installation prefix", "action": "store"},
        "--pkg-format": {
            "help": "format of result binary package",
            "action": "store",
        },
        "--quickpkg-direct": {
            "help": "Enable use of installed packages directly as binary packages",
            "choices": y_or_n,
        },
        "--quickpkg-direct-root": {
            "help": "Specify the root to use as the --quickpkg-direct package source",
            "action": "store",
        },
        "--quiet": {
            "shortopt": "-q",
            "help": "reduced or condensed output",
            "choices": true_y_or_n,
        },
        "--quiet-build": {
            "help": "redirect build output to logs",
            "choices": true_y_or_n,
        },
        "--quiet-fail": {
            "help": "suppresses display of the build log on stdout",
            "choices": true_y_or_n,
        },
        "--read-news": {
            "help": "offer to read unread news via eselect",
            "choices": true_y_or_n,
        },
        "--rebuild-if-new-slot": {
            "help": (
                "Automatically rebuild or reinstall packages when slot/sub-slot := "
                "operator dependencies can be satisfied by a newer slot, so that "
                "older packages slots will become eligible for removal by the "
                "--depclean action as soon as possible."
            ),
            "choices": true_y_or_n,
        },
        "--rebuild-if-new-rev": {
            "help": "Rebuild packages when dependencies that are "
            + "used at both build-time and run-time are built, "
            + "if the dependency is not already installed with the "
            + "same version and revision.",
            "choices": true_y_or_n,
        },
        "--rebuild-if-new-ver": {
            "help": "Rebuild packages when dependencies that are "
            + "used at both build-time and run-time are built, "
            + "if the dependency is not already installed with the "
            + "same version. Revision numbers are ignored.",
            "choices": true_y_or_n,
        },
        "--rebuild-if-unbuilt": {
            "help": "Rebuild packages when dependencies that are "
            + "used at both build-time and run-time are built.",
            "choices": true_y_or_n,
        },
        "--rebuilt-binaries": {
            "help": "replace installed packages with binary "
            + "packages that have been rebuilt",
            "choices": true_y_or_n,
        },
        "--rebuilt-binaries-timestamp": {
            "help": "use only binaries that are newer than this "
            + "timestamp for --rebuilt-binaries",
            "action": "store",
        },
        "--regex-search-auto": {
            "help": "Enable or disable automatic regular expression detection for search actions",
            "choices": y_or_n,
            "default": "y",
        },
        "--root": {
            "help": "specify the target root filesystem for merging packages",
            "action": "store",
        },
        "--root-deps": {
            "help": "modify interpretation of dependencies",
            "choices": ("True", "rdeps"),
        },
        "--search-index": {
            "help": "Enable or disable indexed search (enabled by default)",
            "choices": y_or_n,
        },
        "--search-similarity": {
            "help": (
                "Set minimum similarity percentage for fuzzy seach "
                "(a floating-point number between 0 and 100)"
            ),
            "action": "store",
        },
        "--select": {
            "shortopt": "-w",
            "help": "add specified packages to the world set "
            + "(inverse of --oneshot)",
            "choices": true_y_or_n,
        },
        "--selective": {"help": "identical to --noreplace", "choices": true_y_or_n},
        "--sync-submodule": {
            "help": (
                "Restrict sync to the specified submodule(s)." " (--sync action only)"
            ),
            "choices": tuple(_SUBMODULE_PATH_MAP),
            "action": "append",
        },
        "--sysroot": {
            "help": "specify the location for build dependencies specified in DEPEND",
            "action": "store",
        },
        "--use-ebuild-visibility": {
            "help": "use unbuilt ebuild metadata for visibility checks on built packages",
            "choices": true_y_or_n,
        },
        "--useoldpkg-atoms": {
            "help": "A space separated list of package names or slot atoms. "
            + "Emerge will prefer matching binary packages over newer unbuilt packages. ",
            "action": "append",
        },
        "--usepkg": {
            "shortopt": "-k",
            "help": "use binary packages",
            "choices": true_y_or_n,
        },
        "--usepkgonly": {
            "shortopt": "-K",
            "help": "use only binary packages",
            "choices": true_y_or_n,
        },
        "--usepkg-exclude-live": {
            "help": "do not install from binary packages for live ebuilds",
            "choices": true_y_or_n,
        },
        "--verbose": {
            "shortopt": "-v",
            "help": "verbose output",
            "choices": true_y_or_n,
        },
        "--verbose-slot-rebuilds": {
            "help": "verbose slot rebuild output",
            "choices": true_y_or_n,
        },
        "--with-test-deps": {
            "help": "pull in test deps for packages " + "matched by arguments",
            "choices": true_y_or_n,
        },
    }

    parser = argparse.ArgumentParser(add_help=False)

    for action_opt in actions:
        parser.add_argument(
            f"--{action_opt}",
            action="store_true",
            dest=action_opt.replace("-", "_"),
            default=False,
        )
    for myopt in options:
        parser.add_argument(
            myopt,
            action="store_true",
            dest=myopt.lstrip("--").replace("-", "_"),
            default=False,
        )
    for shortopt, longopt in shortmapping.items():
        parser.add_argument(
            f"-{shortopt}",
            action="store_true",
            dest=longopt.lstrip("--").replace("-", "_"),
            default=False,
        )
    for myalias, myopt in longopt_aliases.items():
        parser.add_argument(
            myalias,
            action="store_true",
            dest=myopt.lstrip("--").replace("-", "_"),
            default=False,
        )

    for myopt, kwargs in argument_options.items():
        shortopt = kwargs.pop("shortopt", None)
        args = [myopt]
        if shortopt is not None:
            args.append(shortopt)
        parser.add_argument(dest=myopt.lstrip("--").replace("-", "_"), *args, **kwargs)

    parser.add_argument("positional_args", nargs="*")

    tmpcmdline = insert_optional_args(tmpcmdline)

    myoptions = getattr(parser, "parse_intermixed_args", parser.parse_args)(
        args=tmpcmdline
    )

    if myoptions.alert in true_y:
        myoptions.alert = True
    else:
        myoptions.alert = None

    if myoptions.ask in true_y:
        myoptions.ask = True
    else:
        myoptions.ask = None

    if myoptions.autounmask in true_y:
        myoptions.autounmask = True

    if myoptions.autounmask_continue in true_y:
        myoptions.autounmask_continue = True

    if myoptions.autounmask_only in true_y:
        myoptions.autounmask_only = True
    else:
        myoptions.autounmask_only = None

    if myoptions.autounmask_unrestricted_atoms in true_y:
        myoptions.autounmask_unrestricted_atoms = True

    if myoptions.autounmask_keep_keywords in true_y:
        myoptions.autounmask_keep_keywords = True

    if myoptions.autounmask_keep_masks in true_y:
        myoptions.autounmask_keep_masks = True

    if myoptions.autounmask_write in true_y:
        myoptions.autounmask_write = True

    if myoptions.binpkg_changed_deps is not None:
        if myoptions.binpkg_changed_deps in true_y:
            myoptions.binpkg_changed_deps = "y"
        else:
            myoptions.binpkg_changed_deps = "n"

    if myoptions.buildpkg in true_y:
        myoptions.buildpkg = True

    if myoptions.buildpkg_exclude:
        bad_atoms = _find_bad_atoms(myoptions.buildpkg_exclude, less_strict=True)
        if bad_atoms and not silent:
            invalid_atoms = ",".join(bad_atoms)
            parser.error(
                f"Invalid Atom(s) in --buildpkg-exclude parameter: '{invalid_atoms}'\n"
            )

    if myoptions.changed_deps is not None:
        if myoptions.changed_deps in true_y:
            myoptions.changed_deps = "y"
        else:
            myoptions.changed_deps = "n"

    if myoptions.changed_deps_report is not None:
        if myoptions.changed_deps_report in true_y:
            myoptions.changed_deps_report = "y"
        else:
            myoptions.changed_deps_report = "n"

    if myoptions.changed_slot is not None:
        if myoptions.changed_slot in true_y:
            myoptions.changed_slot = True
        else:
            myoptions.changed_slot = None

    if myoptions.changed_use is not False:
        myoptions.reinstall = "changed-use"
        myoptions.changed_use = False

    if myoptions.deselect in true_y:
        myoptions.deselect = True

    if myoptions.binpkg_respect_use is not None:
        if myoptions.binpkg_respect_use in true_y:
            myoptions.binpkg_respect_use = "y"
        else:
            myoptions.binpkg_respect_use = "n"

    if myoptions.complete_graph in true_y:
        myoptions.complete_graph = True
    else:
        myoptions.complete_graph = None

    if myoptions.depclean_lib_check in true_y:
        myoptions.depclean_lib_check = True

    candidate_bad_options = (
        (myoptions.exclude, "exclude"),
        (myoptions.reinstall_atoms, "reinstall-atoms"),
        (myoptions.rebuild_exclude, "rebuild-exclude"),
        (myoptions.rebuild_ignore, "rebuild-ignore"),
        (myoptions.usepkg_exclude, "usepkg-exclude"),
        (myoptions.useoldpkg_atoms, "useoldpkg-atoms"),
    )
    bad_options = (
        (_find_bad_atoms(atoms), flag) for atoms, flag in candidate_bad_options if atoms
    )

    for bad_atoms, flag in bad_options:
        if bad_atoms and not silent:
            invalid_atoms = ",".join(bad_atoms)
            parser.error(
                f"Invalid Atom(s) in --{flag} parameter: '{invalid_atoms}' (only package names and slot atoms (with wildcards) allowed)\n"
            )

    if myoptions.fail_clean in true_y:
        myoptions.fail_clean = True

    if myoptions.fuzzy_search in true_y:
        myoptions.fuzzy_search = True

    if myoptions.getbinpkg in true_y:
        myoptions.getbinpkg = True
    else:
        myoptions.getbinpkg = None

    if myoptions.getbinpkgonly in true_y:
        myoptions.getbinpkgonly = True
    else:
        myoptions.getbinpkgonly = None

    if myoptions.ignore_world in true_y:
        myoptions.ignore_world = True

    if myoptions.keep_going in true_y:
        myoptions.keep_going = True
    else:
        myoptions.keep_going = None

    if myoptions.package_moves in true_y:
        myoptions.package_moves = True

    if myoptions.quiet in true_y:
        myoptions.quiet = True
    else:
        myoptions.quiet = None

    if myoptions.quiet_build in true_y:
        myoptions.quiet_build = "y"

    if myoptions.quiet_fail in true_y:
        myoptions.quiet_fail = "y"

    if myoptions.read_news in true_y:
        myoptions.read_news = True
    else:
        myoptions.read_news = None

    if myoptions.rebuild_if_new_slot in true_y:
        myoptions.rebuild_if_new_slot = "y"

    if myoptions.rebuild_if_new_ver in true_y:
        myoptions.rebuild_if_new_ver = True
    else:
        myoptions.rebuild_if_new_ver = None

    if myoptions.rebuild_if_new_rev in true_y:
        myoptions.rebuild_if_new_rev = True
        myoptions.rebuild_if_new_ver = None
    else:
        myoptions.rebuild_if_new_rev = None

    if myoptions.rebuild_if_unbuilt in true_y:
        myoptions.rebuild_if_unbuilt = True
        myoptions.rebuild_if_new_rev = None
        myoptions.rebuild_if_new_ver = None
    else:
        myoptions.rebuild_if_unbuilt = None

    if myoptions.rebuilt_binaries in true_y:
        myoptions.rebuilt_binaries = True

    if myoptions.root_deps in true_y:
        myoptions.root_deps = True

    if myoptions.select in true_y:
        myoptions.select = True
        myoptions.oneshot = False
    elif myoptions.select == "n":
        myoptions.oneshot = True

    if myoptions.selective in true_y:
        myoptions.selective = True

    if myoptions.backtrack is not None:
        try:
            backtrack = int(myoptions.backtrack)
        except (OverflowError, ValueError):
            backtrack = -1

        if backtrack < 0:
            backtrack = None
            if not silent:
                parser.error(
                    f"Invalid --backtrack parameter: '{myoptions.backtrack}'\n"
                )

        myoptions.backtrack = backtrack

    if myoptions.deep is not None:
        deep = None
        if myoptions.deep == "True":
            deep = True
        else:
            try:
                deep = int(myoptions.deep)
            except (OverflowError, ValueError):
                deep = -1

        if deep is not True and deep < 0:
            deep = None
            if not silent:
                parser.error(f"Invalid --deep parameter: '{myoptions.deep}'\n")

        myoptions.deep = deep

    if myoptions.jobs is not None:
        jobs = None
        if myoptions.jobs == "True":
            jobs = True
        else:
            try:
                jobs = int(myoptions.jobs)
            except ValueError:
                jobs = -1

        if jobs is not True and jobs < 1:
            jobs = None
            if not silent:
                parser.error(f"Invalid --jobs parameter: '{myoptions.jobs}'\n")

        myoptions.jobs = jobs

    if myoptions.load_average == "True":
        myoptions.load_average = None

    if myoptions.load_average:
        try:
            load_average = float(myoptions.load_average)
        except ValueError:
            load_average = 0.0

        if load_average <= 0.0:
            load_average = None
            if not silent:
                parser.error(
                    f"Invalid --load-average parameter: '{myoptions.load_average}'\n"
                )

        myoptions.load_average = load_average

    if myoptions.rebuilt_binaries_timestamp:
        try:
            rebuilt_binaries_timestamp = int(myoptions.rebuilt_binaries_timestamp)
        except ValueError:
            rebuilt_binaries_timestamp = -1

        if rebuilt_binaries_timestamp < 0:
            rebuilt_binaries_timestamp = 0
            if not silent:
                parser.error(
                    f"Invalid --rebuilt-binaries-timestamp parameter: '{myoptions.rebuilt_binaries_timestamp}'\n"
                )

        myoptions.rebuilt_binaries_timestamp = rebuilt_binaries_timestamp

    if myoptions.search_similarity:
        try:
            search_similarity = float(myoptions.search_similarity)
        except ValueError:
            parser.error(
                f"Invalid --search-similarity parameter (not a number): '{myoptions.search_similarity}'\n"
            )

        if search_similarity < 0 or search_similarity > 100:
            parser.error(
                f"Invalid --search-similarity parameter (not between 0 and 100): '{myoptions.search_similarity}'\n"
            )

        myoptions.search_similarity = search_similarity

    if myoptions.use_ebuild_visibility in true_y:
        myoptions.use_ebuild_visibility = True
    else:
        # None or "n"
        pass

    if myoptions.usepkg in true_y:
        myoptions.usepkg = True
    else:
        myoptions.usepkg = None

    if myoptions.usepkgonly in true_y:
        myoptions.usepkgonly = True
    else:
        myoptions.usepkgonly = None

    if myoptions.usepkg_exclude_live in true_y:
        myoptions.usepkg_exclude_live = True
    else:
        myoptions.usepkg_exclude_live = None

    if myoptions.verbose in true_y:
        myoptions.verbose = True
    else:
        myoptions.verbose = None

    if myoptions.with_test_deps in true_y:
        myoptions.with_test_deps = True
    else:
        myoptions.with_test_deps = None

    for myopt in options:
        v = getattr(myoptions, myopt.lstrip("--").replace("-", "_"))
        if v:
            myopts[myopt] = True

    for myopt in argument_options:
        v = getattr(myoptions, myopt.lstrip("--").replace("-", "_"), None)
        if v is not None:
            myopts[myopt] = v

    if myoptions.searchdesc:
        myoptions.search = True

    for action_opt in actions:
        v = getattr(myoptions, action_opt.replace("-", "_"))
        if v:
            if myaction:
                multiple_actions(myaction, action_opt)
                sys.exit(1)
            myaction = action_opt

    if myaction is None and myoptions.deselect is True:
        myaction = "deselect"

    return myaction, myopts, myoptions.positional_args


def profile_check(trees, myaction):
    if myaction in ("help", "info", "search", "sync", "version"):
        return os.EX_OK
    for root_trees in trees.values():
        if (
            root_trees["root_config"].settings.profiles
            and "ARCH" in root_trees["root_config"].settings
        ):
            continue
        # generate some profile related warning messages
        validate_ebuild_environment(trees)
        msg = (
            "Your current profile is invalid. If you have just changed "
            "your profile configuration, you should revert back to the "
            "previous configuration. Allowed actions are limited to "
            "--help, --info, --search, --sync, and --version."
        )
        writemsg_level(
            "".join(f"!!! {l}\n" for l in textwrap.wrap(msg, 70)),
            level=logging.ERROR,
            noiselevel=-1,
        )
        return 1
    return os.EX_OK


def emerge_main(args: Optional[list[str]] = None):
    """
    Entry point of emerge

    Processes command line arguments (default: sys.argv[1:]) and decides
    what the current run of emerge should by creating `emerge_config`
    """
    if args is None:
        args = sys.argv[1:]

    args = portage._decode_argv(args)

    # Use system locale.
    try:
        locale.setlocale(locale.LC_ALL, "")
    except locale.Error as e:
        writemsg_level(f"setlocale: {e}\n", level=logging.WARN)

    # Disable color until we're sure that it should be enabled (after
    # EMERGE_DEFAULT_OPTS has been parsed).
    portage.output.havecolor = 0

    # This first pass is just for options that need to be known as early as
    # possible, such as --config-root.  They will be parsed again later,
    # together with EMERGE_DEFAULT_OPTS (which may vary depending on the
    # the value of --config-root).
    myaction, myopts, myfiles = parse_opts(args, silent=True)
    if "--debug" in myopts:
        os.environ["PORTAGE_DEBUG"] = "1"
        portage.util.initialize_logger(logging.DEBUG)
    else:
        portage.util.initialize_logger()

    if "--config-root" in myopts:
        os.environ["PORTAGE_CONFIGROOT"] = myopts["--config-root"]
    if "--sysroot" in myopts:
        os.environ["SYSROOT"] = myopts["--sysroot"]
    if "--root" in myopts:
        os.environ["ROOT"] = myopts["--root"]
    if "--prefix" in myopts:
        os.environ["EPREFIX"] = myopts["--prefix"]
    if "--accept-properties" in myopts:
        os.environ["ACCEPT_PROPERTIES"] = myopts["--accept-properties"]
    if "--accept-restrict" in myopts:
        os.environ["ACCEPT_RESTRICT"] = myopts["--accept-restrict"]

    # optimize --help (no need to load config / EMERGE_DEFAULT_OPTS)
    if myaction == "help":
        emerge_help()
        return os.EX_OK
    if myaction == "moo":
        print(COWSAY_MOO.format(platform.system()))
        return os.EX_OK
    if myaction == "sync":
        # need to set this to True now in order for the repository config
        # loading to allow new repos with non-existent directories
        portage._sync_mode = True

    # Verify that /dev/null exists and is a device file as a cheap early
    # filter for obviously broken /dev/s.
    try:
        if os.stat(os.devnull).st_rdev == 0:
            writemsg_level(
                "Failed to validate a sane '/dev'.\n"
                "'/dev/null' is not a device file.\n",
                level=logging.ERROR,
                noiselevel=-1,
            )
            return 1
    except OSError:
        writemsg_level(
            "Failed to validate a sane '/dev'.\n" "'/dev/null' does not exist.\n",
            level=logging.ERROR,
            noiselevel=-1,
        )
        return 1

    # Verify that BASH process substitution works as another cheap early
    # filter. Process substitution uses '/dev/fd'.
    with open(os.devnull, "r+b") as dev_null:
        fd_pipes = {
            0: dev_null.fileno(),
            1: dev_null.fileno(),
            2: dev_null.fileno(),
        }
        exit_code = portage.process.spawn_bash(
            "[[ $(< <(echo foo) ) == foo ]]", fd_pipes=fd_pipes
        )
        if exit_code != 0:
            writemsg_level(
                "Failed to validate a sane '/dev'.\n"
                "bash process substitution doesn't work; this may be an "
                "indication of a broken '/dev/fd'.\n",
                level=logging.ERROR,
                noiselevel=-1,
            )
            return 1

    # Portage needs to ensure a sane umask for the files it creates.
    os.umask(0o22)
    emerge_config = load_emerge_config(action=myaction, args=myfiles, opts=myopts)

    # Make locale variables from configuration files (make.defaults, make.conf) affect locale of emerge process.
    for locale_var_name in (
        "LANGUAGE",
        "LC_ALL",
        "LC_ADDRESS",
        "LC_COLLATE",
        "LC_CTYPE",
        "LC_IDENTIFICATION",
        "LC_MEASUREMENT",
        "LC_MESSAGES",
        "LC_MONETARY",
        "LC_NAME",
        "LC_NUMERIC",
        "LC_PAPER",
        "LC_TELEPHONE",
        "LC_TIME",
        "LANG",
    ):
        locale_var_value = emerge_config.running_config.settings.get(locale_var_name)
        if locale_var_value is not None:
            os.environ.setdefault(locale_var_name, locale_var_value)
    try:
        locale.setlocale(locale.LC_ALL, "")
    except locale.Error as e:
        writemsg_level(f"setlocale: {e}\n", level=logging.WARN)

    tmpcmdline = []
    if "--ignore-default-opts" not in myopts:
        tmpcmdline.extend(
            portage.util.shlex_split(
                emerge_config.target_config.settings.get("EMERGE_DEFAULT_OPTS", "")
            )
        )
    tmpcmdline.extend(args)
    emerge_config.action, emerge_config.opts, emerge_config.args = parse_opts(
        tmpcmdline
    )

    try:
        return run_action(emerge_config)
    finally:
        # Call destructors for our portdbapi instances.
        for x in emerge_config.trees.values():
            if "porttree" in x.lazy_items:
                continue
            x["porttree"].dbapi.close_caches()



# Copyright 2010-2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import platform
import pty
import termios

from portage import os
from portage.output import get_term_size, set_term_size
from portage.util import writemsg

# Disable the use of openpty on Solaris as it seems Python's openpty
# implementation doesn't play nice on Solaris with Portage's
# behaviour causing hangs/deadlocks.
# Additional note for the future: on Interix, pipes do NOT work, so
# _disable_openpty on Interix must *never* be True
_disable_openpty = platform.system() in ("SunOS",)

_fbsd_test_pty = platform.system() == "FreeBSD"


def _create_pty_or_pipe(copy_term_size=None):
    """
    Try to create a pty and if then fails then create a normal
    pipe instead.

    @param copy_term_size: If a tty file descriptor is given
            then the term size will be copied to the pty.
    @type copy_term_size: int
    @rtype: tuple
    @return: A tuple of (is_pty, master_fd, slave_fd) where
            is_pty is True if a pty was successfully allocated, and
            False if a normal pipe was allocated.
    """

    got_pty = False

    global _disable_openpty, _fbsd_test_pty

    if _fbsd_test_pty and not _disable_openpty:
        # Test for python openpty breakage after freebsd7 to freebsd8
        # upgrade, which results in a 'Function not implemented' error
        # and the process being killed.
        pid = os.fork()
        if pid == 0:
            pty.openpty()
            os._exit(os.EX_OK)
        pid, status = os.waitpid(pid, 0)
        if (status & 0xFF) == 140:
            _disable_openpty = True
        _fbsd_test_pty = False

    if _disable_openpty:
        master_fd, slave_fd = os.pipe()
    else:
        try:
            master_fd, slave_fd = pty.openpty()
            got_pty = True
        except OSError as e:
            _disable_openpty = True
            writemsg(f"openpty failed: '{str(e)}'\n", noiselevel=-1)
            del e
            master_fd, slave_fd = os.pipe()

    if got_pty:
        # Disable post-processing of output since otherwise weird
        # things like \n -> \r\n transformations may occur.
        mode = termios.tcgetattr(slave_fd)
        mode[1] &= ~termios.OPOST
        termios.tcsetattr(slave_fd, termios.TCSANOW, mode)

    if got_pty and copy_term_size is not None and os.isatty(copy_term_size):
        rows, columns = get_term_size()
        set_term_size(rows, columns, slave_fd)

    return (got_pty, master_fd, slave_fd)
